import shutil

from mayan.apps.documents.literals import DOCUMENT_FILE_ACTION_PAGES_NEW
from mayan.apps.documents.tests.literals import (
    TEST_DOCUMENT_DESCRIPTION, TEST_SMALL_DOCUMENT_PATH
)
from mayan.apps.storage.utils import fs_cleanup, mkdtemp

from ..literals import SOURCE_CHOICE_WEB_FORM, SOURCE_UNCOMPRESS_CHOICE_Y
from ..models.staging_folder_sources import StagingFolderSource
from ..models.watch_folder_sources import WatchFolderSource
from ..models.webform_sources import WebFormSource

from .literals import (
    TEST_SOURCE_LABEL, TEST_SOURCE_LABEL_EDITED, TEST_SOURCE_UNCOMPRESS_N,
    TEST_STAGING_PREVIEW_WIDTH
)


class DocumentUploadIssueTestMixin:
    def _request_test_source_create_view(self):
        return self.post(
            viewname='sources:setup_source_create', kwargs={
                'source_type_name': SOURCE_CHOICE_WEB_FORM
            }, data={
                'enabled': True, 'label': 'test', 'uncompress': 'n'
            }
        )

    def _request_test_source_edit_view(self):
        return self.post(
            viewname='documents:document_properties_edit', kwargs={
                'document_id': self.test_document.pk
            },
            data={
                'description': TEST_DOCUMENT_DESCRIPTION,
                'label': self.test_document.label,
                'language': self.test_document.language
            }
        )


class DocumentUploadWizardViewTestMixin:
    def _request_upload_wizard_view(self, document_path=TEST_SMALL_DOCUMENT_PATH):
        with open(file=document_path, mode='rb') as file_object:
            return self.post(
                viewname='sources:document_upload_interactive', kwargs={
                    'source_id': self.test_source.pk
                }, data={
                    'source-file': file_object,
                    'document_type_id': self.test_document_type.pk,
                }
            )

    def _request_upload_interactive_view(self):
        return self.get(
            viewname='sources:document_upload_interactive', data={
                'document_type_id': self.test_document_type.pk,
            }
        )


class DocumentFileUploadViewTestMixin:
    def _request_document_file_upload_view(self):
        with open(file=TEST_SMALL_DOCUMENT_PATH, mode='rb') as file_object:
            return self.post(
                viewname='sources:document_file_upload', kwargs={
                    'document_id': self.test_document.pk,
                    'source_id': self.test_source.pk,
                }, data={
                    'document-action': DOCUMENT_FILE_ACTION_PAGES_NEW,
                    'source-file': file_object
                }
            )

    def _request_document_file_upload_no_source_view(self):
        with open(file=TEST_SMALL_DOCUMENT_PATH, mode='rb') as file_object:
            return self.post(
                viewname='sources:document_file_upload', kwargs={
                    'document_id': self.test_document.pk,
                }, data={
                    'document-action': DOCUMENT_FILE_ACTION_PAGES_NEW,
                    'source-file': file_object
                }
            )


class StagingFolderAPIViewTestMixin:
    def setUp(self):
        super().setUp()
        self.test_staging_folders = []

    def tearDown(self):
        for test_staging_folder in self.test_staging_folders:
            fs_cleanup(filename=test_staging_folder.folder_path)
            self.test_staging_folders.remove(test_staging_folder)

        super().tearDown()

    def _request_test_staging_folder_create_api_view(self):
        return self.post(
            viewname='rest_api:stagingfolder-list', data={
                'label': TEST_SOURCE_LABEL,
                'folder_path': mkdtemp(),
                'preview_width': TEST_STAGING_PREVIEW_WIDTH,
                'uncompress': TEST_SOURCE_UNCOMPRESS_N,
            }
        )

        self.test_staging_folder = StagingFolderSource.objects.first()
        self.test_staging_folders.append(self.test_staging_folder)

    def _request_test_staging_folder_delete_api_view(self):
        return self.delete(
            viewname='rest_api:stagingfolder-detail', kwargs={
                'pk': self.test_staging_folder.pk
            }
        )

    def _request_test_staging_folder_edit_api_view(self, extra_data=None, verb='patch'):
        data = {
            'label': TEST_SOURCE_LABEL_EDITED,
        }

        if extra_data:
            data.update(extra_data)

        return getattr(self, verb)(
            viewname='rest_api:stagingfolder-detail', kwargs={
                'pk': self.test_staging_folder.pk
            }, data=data
        )

    def _request_test_staging_folder_list_api_view(self):
        return self.get(viewname='rest_api:stagingfolder-list')


class StagingFolderFileAPIViewTestMixin:
    def _request_test_staging_folder_file_delete_api_view(self):
        return self.delete(
            viewname='rest_api:stagingfolderfile-detail', kwargs={
                'staging_folder_pk': self.test_staging_folder.pk,
                'encoded_filename': self.test_staging_folder_file.encoded_filename
            }
        )

    def _request_test_staging_folder_file_detail_api_view(self):
        return self.get(
            viewname='rest_api:stagingfolderfile-detail', kwargs={
                'staging_folder_pk': self.test_staging_folder.pk,
                'encoded_filename': self.test_staging_folder_file.encoded_filename
            }
        )

    def _request_test_staging_folder_file_upload_api_view(self):
        return self.post(
            viewname='rest_api:stagingfolderfile-upload', kwargs={
                'staging_folder_pk': self.test_staging_folder.pk,
                'encoded_filename': self.test_staging_folder_file.encoded_filename
            }, data={'document_type': self.test_document_type.pk}
        )


class StagingFolderTestMixin:
    def setUp(self):
        super().setUp()
        self.test_staging_folders = []

    def tearDown(self):
        for test_staging_folder in self.test_staging_folders:
            fs_cleanup(filename=test_staging_folder.folder_path)
            self.test_staging_folders.remove(test_staging_folder)

        super().tearDown()

    def _create_test_staging_folder(self):
        self.test_staging_folder = StagingFolderSource.objects.create(
            label=TEST_SOURCE_LABEL,
            folder_path=mkdtemp(),
            preview_width=TEST_STAGING_PREVIEW_WIDTH,
            uncompress=TEST_SOURCE_UNCOMPRESS_N,
        )
        self.test_staging_folders.append(self.test_staging_folder)

    def _copy_test_document(self):
        shutil.copy(
            src=TEST_SMALL_DOCUMENT_PATH,
            dst=self.test_staging_folder.folder_path
        )
        self.test_staging_folder_file = list(
            self.test_staging_folder.get_files()
        )[0]


class StagingFolderViewTestMixin:
    def _request_test_staging_file_delete_view(self, staging_folder, staging_file):
        return self.post(
            viewname='sources:staging_file_delete', kwargs={
                'staging_folder_id': staging_folder.pk,
                'encoded_filename': staging_file.encoded_filename
            }
        )


class SourceTestMixin:
    auto_create_test_source = True

    def setUp(self):
        super().setUp()
        if self.auto_create_test_source:
            self._create_test_source()

    def _create_test_source(self):
        self.test_source = WebFormSource.objects.create(
            enabled=True, label=TEST_SOURCE_LABEL,
            uncompress=TEST_SOURCE_UNCOMPRESS_N
        )


class SourceViewTestMixin:
    def _request_setup_source_list_view(self):
        return self.get(viewname='sources:setup_source_list')

    def _request_setup_source_check_get_view(self):
        return self.get(
            viewname='sources:setup_source_check', kwargs={
                'source_id': self.test_source.pk
            }
        )

    def _request_setup_source_check_post_view(self):
        return self.post(
            viewname='sources:setup_source_check', kwargs={
                'source_id': self.test_source.pk
            }
        )

    def _request_setup_source_create_view(self):
        return self.post(
            kwargs={
                'source_type_name': SOURCE_CHOICE_WEB_FORM
            }, viewname='sources:setup_source_create', data={
                'enabled': True, 'label': TEST_SOURCE_LABEL,
                'uncompress': TEST_SOURCE_UNCOMPRESS_N
            }
        )

    def _request_setup_source_delete_view(self):
        return self.post(
            viewname='sources:setup_source_delete', kwargs={
                'source_id': self.test_source.pk
            }
        )

    def _request_setup_source_edit_view(self):
        return self.post(
            viewname='sources:setup_source_edit', kwargs={
                'source_id': self.test_source.pk
            }, data={
                'label': TEST_SOURCE_LABEL_EDITED,
                'uncompress': self.test_source.uncompress
            }
        )


class WatchFolderTestMixin:
    def setUp(self):
        super().setUp()
        self.temporary_directory = mkdtemp()
        self.test_watch_folders = []

    def tearDown(self):
        shutil.rmtree(self.temporary_directory)
        super().tearDown()

    def _create_test_watchfolder(self, label=None):
        kwargs = dict(
            document_type=self.test_document_type,
            folder_path=self.temporary_directory,
            include_subdirectories=False,
            label=label or 'test source',
            uncompress=SOURCE_UNCOMPRESS_CHOICE_Y
        )

        self.test_watch_folder = WatchFolderSource.objects.create(**kwargs)
        self.test_watch_folders.append(self.test_watch_folder)



from django.db import migrations, reset_queries


def operation_document_version_page_create(apps, schema_editor):
    ContentType = apps.get_model(
        app_label='contenttypes', model_name='ContentType'
    )
    DocumentFilePage = apps.get_model(
        app_label='documents', model_name='DocumentFilePage'
    )
    DocumentVersion = apps.get_model(
        app_label='documents', model_name='DocumentVersion'
    )

    content_type = ContentType.objects.get_for_model(model=DocumentFilePage)
    content_type_id = content_type.id
    cursor_main = schema_editor.connection.create_cursor(name='document_file_to_version')
    cursor_document_version_page = schema_editor.connection.cursor()

    query = '''
    SELECT
        {documents_documentfile}.{document_id},
        {documents_documentfilepage}.{document_file_id},
        {documents_documentfilepage}.{id}
    FROM {documents_documentfilepage}
    INNER JOIN {documents_documentfile} ON (
        {documents_documentfilepage}.{document_file_id} = {documents_documentfile}.{id}
    ) ORDER BY {documents_documentfilepage}.{id} ASC
    '''.format(
        documents_documentfile=schema_editor.connection.ops.quote_name('documents_documentfile'),
        document_id=schema_editor.connection.ops.quote_name('document_id'),
        documents_documentfilepage=schema_editor.connection.ops.quote_name('documents_documentfilepage'),
        document_file_id=schema_editor.connection.ops.quote_name('document_file_id'),
        id=schema_editor.connection.ops.quote_name('id')
    )

    cursor_main.execute(query)

    class DummyDocumentVersion:
        def save(self):
            """Does not do anything."""

    document_file_id_last = None
    document_id_last = None
    document_version = DummyDocumentVersion()

    document_version_page_insert_query = '''
        INSERT INTO documents_documentversionpage (
            document_version_id,content_type_id,object_id,page_number
        ) VALUES {};
    '''
    document_version_page_values = []
    page_number = 1

    FETCH_SIZE = 100000

    while True:
        rows = cursor_main.fetchmany(FETCH_SIZE)

        if not rows:
            break

        for row in rows:
            document_id, document_file_id, document_file_page_id = row

            if document_id_last != document_id:
                document_version.active = True
                document_version.save()
                document_id_last = document_id

            if document_file_id_last != document_file_id:
                document_version = DocumentVersion.objects.create(document_id=document_id)
                document_version_id = document_version.pk
                document_file_id_last = document_file_id
                if document_version_page_values:
                    final_query = document_version_page_insert_query.format(
                        ('(%s,%s,%s,%s),' * (page_number - 1))[:-1]
                    )

                    page_number = 1

                    cursor_document_version_page.execute(
                        final_query, document_version_page_values
                    )
                    reset_queries()

                document_version_page_values = []

            document_version_page_values += (
                document_version_id, content_type_id, document_file_page_id,
                page_number
            )

            page_number += 1

    if page_number > 1:
        final_query = document_version_page_insert_query.format(
            ('(%s,%s,%s,%s),' * (page_number - 1))[:-1]
        )
        cursor_document_version_page.execute(
            final_query, document_version_page_values
        )


def operation_document_version_page_create_reverse(apps, schema_editor):
    DocumentVersion = apps.get_model(
        app_label='documents', model_name='DocumentVersion'
    )

    DocumentVersion.objects.using(schema_editor.connection.alias).all().delete()


class Migration(migrations.Migration):
    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('documents', '0062_auto_20200920_0614'),
    ]

    operations = [
        migrations.RunPython(
            code=operation_document_version_page_create,
            reverse_code=operation_document_version_page_create_reverse
        ),
    ]



from mayan.apps.testing.tests.base import GenericViewTestCase

from .mixins import MessageTestMixin, MessageViewTestMixin

from ..events import event_message_created, event_message_edited
from ..models import Message
from ..permissions import (
    permission_message_create, permission_message_delete,
    permission_message_view
)


class MessageViewTestCase(
    MessageTestMixin, MessageViewTestMixin, GenericViewTestCase
):
    def test_message_create_view_no_permission(self):
        message_count = Message.objects.count()

        self._clear_events()

        response = self._request_test_message_create_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(Message.objects.count(), message_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_message_create_view_with_permissions(self):
        self.grant_permission(permission=permission_message_create)

        message_count = Message.objects.count()

        self._clear_events()

        response = self._request_test_message_create_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Message.objects.count(), message_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_message)
        self.assertEqual(events[0].verb, event_message_created.id)

    def test_message_delete_view_no_permission(self):
        self._create_test_message()

        message_count = Message.objects.count()

        self._clear_events()

        response = self._request_test_message_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Message.objects.count(), message_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_message_delete_view_with_access(self):
        self._create_test_message()

        self.grant_access(
            obj=self.test_message, permission=permission_message_delete
        )

        message_count = Message.objects.count()

        self._clear_events()

        response = self._request_test_message_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Message.objects.count(), message_count - 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_message_detail_view_no_permission(self):
        self._create_test_message()

        test_message_read = self.test_message.read

        self._clear_events()

        response = self._request_test_message_detail_view()
        self.assertEqual(response.status_code, 404)

        self.test_message.refresh_from_db()
        self.assertEqual(
            self.test_message.read, test_message_read
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_message_detail_view_with_access(self):
        self._create_test_message()

        self.grant_access(
            obj=self.test_message, permission=permission_message_view
        )

        test_message_read = self.test_message.read

        self._clear_events()

        response = self._request_test_message_detail_view()
        self.assertContains(
            text=self.test_message.get_rendered_body(), response=response,
            status_code=200
        )

        self.test_message.refresh_from_db()
        self.assertNotEqual(
            self.test_message.read, test_message_read
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_message)
        self.assertEqual(events[0].verb, event_message_edited.id)

    def test_message_list_view_with_no_permission(self):
        self._create_test_message()

        self._clear_events()

        response = self._request_test_message_list_view()
        self.assertNotContains(
            response=response, text=self.test_message.subject, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_message_list_view_with_access(self):
        self._create_test_message()

        self.grant_access(
            obj=self.test_message, permission=permission_message_view
        )

        self._clear_events()

        response = self._request_test_message_list_view()
        self.assertContains(
            response=response, text=self.test_message.subject, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.smart_settings.classes import SettingNamespace

from .literals import DEFAULT_REST_API_DISABLE_LINKS

namespace = SettingNamespace(
    label=_('REST API'), name='rest_api', version='0001'
)

setting_disable_links = namespace.add_setting(
    default=DEFAULT_REST_API_DISABLE_LINKS,
    global_name='REST_API_DISABLE_LINKS', help_text=_(
        'Disable the REST API links in the tools menu.'
    )
)



import logging

from django.contrib import messages
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _, ungettext

from mayan.apps.common.classes import ModelQueryFields
from mayan.apps.views.generics import (
    MultipleObjectFormActionView, SingleObjectDetailView,
    SingleObjectEditView, SingleObjectListView
)

from ..events import event_document_viewed
from ..forms.document_forms import DocumentForm, DocumentPropertiesForm
from ..forms.document_type_forms import DocumentTypeFilteredSelectForm
from ..icons import icon_document_list
from ..models.document_models import Document
from ..permissions import (
    permission_document_properties_edit, permission_document_view
)

from .document_version_views import DocumentVersionPreviewView

__all__ = (
    'DocumentListView', 'DocumentTypeChangeView',
    'DocumentPropertiesEditView', 'DocumentPreviewView'
)
logger = logging.getLogger(name=__name__)


class DocumentListView(SingleObjectListView):
    object_permission = permission_document_view

    def get_context_data(self, **kwargs):
        try:
            return super().get_context_data(**kwargs)
        except Exception as exception:
            messages.error(
                message=_(
                    'Error retrieving document list: %(exception)s.'
                ) % {
                    'exception': exception
                }, request=self.request
            )
            self.object_list = Document.valid.none()
            return super().get_context_data(**kwargs)

    def get_document_queryset(self):
        return Document.valid.all()

    def get_extra_context(self):
        return {
            'hide_links': True,
            'hide_object': True,
            'list_as_items': True,
            'no_results_icon': icon_document_list,
            'no_results_text': _(
                'This could mean that no documents have been uploaded or '
                'that your user account has not been granted the view '
                'permission for any document or document type.'
            ),
            'no_results_title': _('No documents available'),
            'title': _('All documents'),
        }

    def get_source_queryset(self):
        queryset = ModelQueryFields.get(model=Document).get_queryset()
        return self.get_document_queryset().filter(pk__in=queryset)


class DocumentTypeChangeView(MultipleObjectFormActionView):
    form_class = DocumentTypeFilteredSelectForm
    object_permission = permission_document_properties_edit
    pk_url_kwarg = 'document_id'
    source_queryset = Document.valid
    success_message = _(
        'Document type change request performed on %(count)d document'
    )
    success_message_plural = _(
        'Document type change request performed on %(count)d documents'
    )

    def get_extra_context(self):
        queryset = self.object_list

        result = {
            'submit_label': _('Change'),
            'title': ungettext(
                singular='Change the type of the selected document',
                plural='Change the type of the selected documents',
                number=queryset.count()
            )
        }

        if queryset.count() == 1:
            result.update(
                {
                    'object': queryset.first(),
                    'title': _(
                        'Change the type of the document: %s'
                    ) % queryset.first()
                }
            )

        return result

    def get_form_extra_kwargs(self):
        result = {
            'user': self.request.user
        }

        return result

    def object_action(self, form, instance):
        instance.document_type_change(
            document_type=form.cleaned_data['document_type'],
            _user=self.request.user
        )

        messages.success(
            message=_(
                'Document type for "%s" changed successfully.'
            ) % instance, request=self.request
        )


class DocumentPreviewView(DocumentVersionPreviewView):
    object_permission = permission_document_view
    pk_url_kwarg = 'document_id'
    source_queryset = Document.valid

    def dispatch(self, request, *args, **kwargs):
        result = super(
            DocumentVersionPreviewView, self
        ).dispatch(request=request, *args, **kwargs)
        self.object.add_as_recent_document_for_user(user=request.user)
        event_document_viewed.commit(
            actor=request.user, target=self.object
        )

        return result

    def get_extra_context(self):
        return {
            'hide_labels': True,
            'object': self.object,
            'title': _('Preview of document: %s') % self.object,
        }


class DocumentPropertiesEditView(SingleObjectEditView):
    form_class = DocumentForm
    object_permission = permission_document_properties_edit
    pk_url_kwarg = 'document_id'
    source_queryset = Document.valid

    def dispatch(self, request, *args, **kwargs):
        result = super().dispatch(request, *args, **kwargs)
        self.object.add_as_recent_document_for_user(user=request.user)
        return result

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _('Edit properties of document: %s') % self.object,
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }

    def get_post_action_redirect(self):
        return reverse(
            viewname='documents:document_properties', kwargs={
                'document_id': self.object.pk
            }
        )


class DocumentPropertiesView(SingleObjectDetailView):
    form_class = DocumentPropertiesForm
    object_permission = permission_document_view
    pk_url_kwarg = 'document_id'
    source_queryset = Document.valid

    def dispatch(self, request, *args, **kwargs):
        result = super().dispatch(request, *args, **kwargs)
        self.object.add_as_recent_document_for_user(request.user)
        return result

    def get_extra_context(self):
        return {
            'document': self.object,
            'object': self.object,
            'title': _('Properties of document: %s') % self.object,
        }



import logging

from django.shortcuts import reverse
from django.template import RequestContext
from django.urls import reverse_lazy
from django.utils.translation import ugettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.common.classes import ModelQueryFields
from mayan.apps.documents.models import Document
from mayan.apps.documents.views.document_views import DocumentListView
from mayan.apps.views.generics import (
    MultipleObjectFormActionView, MultipleObjectConfirmActionView,
    SingleObjectCreateView, SingleObjectEditView, SingleObjectListView
)
from mayan.apps.views.mixins import ExternalObjectViewMixin

from .forms import TagMultipleSelectionForm
from .icons import icon_menu_tags, icon_document_tag_remove_submit
from .links import link_document_tag_multiple_attach, link_tag_create
from .models import DocumentTag, Tag
from .permissions import (
    permission_tag_attach, permission_tag_create, permission_tag_delete,
    permission_tag_edit, permission_tag_remove, permission_tag_view
)

logger = logging.getLogger(name=__name__)


class TagAttachActionView(MultipleObjectFormActionView):
    form_class = TagMultipleSelectionForm
    object_permission = permission_tag_attach
    pk_url_kwarg = 'document_id'
    source_queryset = Document.valid
    success_message_single = _(
        'Tags attached to document "%(object)s" successfully.'
    )
    success_message_singular = _(
        'Tags attached to %(count)d document successfully.'
    )
    success_message_plural = _(
        'Tags attached to %(count)d documents successfully.'
    )
    title_single = _('Attach tags to document: %(object)s')
    title_singular = _('Attach tags to %(count)d document.')
    title_plural = _('Attach tags to %(count)d documents.')

    def get_extra_context(self):
        context = {
            'submit_label': _('Attach'),
        }

        if self.object_list.count() == 1:
            context.update(
                {
                    'object': self.object_list.first(),
                }
            )

        return context

    def get_form_extra_kwargs(self):
        kwargs = {
            'help_text': _('Tags to be attached.'),
            'permission': permission_tag_attach,
            'queryset': Tag.objects.all(),
            'user': self.request.user
        }

        if self.object_list.count() == 1:
            kwargs.update(
                {
                    'queryset': Tag.objects.exclude(
                        pk__in=self.object_list.first().tags.all()
                    )
                }
            )

        return kwargs

    def get_post_action_redirect(self):
        if self.object_list.count() == 1:
            return reverse(
                viewname='tags:document_tag_list', kwargs={
                    'document_id': self.object_list.first().pk
                }
            )
        else:
            return super().get_post_action_redirect()

    def object_action(self, form, instance):
        for tag in form.cleaned_data['tags']:
            AccessControlList.objects.check_access(
                obj=tag, permissions=(permission_tag_attach,),
                user=self.request.user
            )

            tag._event_actor = self.request.user
            tag.attach_to(document=instance)


class TagCreateView(SingleObjectCreateView):
    extra_context = {'title': _('Create tag')}
    fields = ('label', 'color')
    model = Tag
    post_action_redirect = reverse_lazy(viewname='tags:tag_list')
    view_permission = permission_tag_create

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class TagDeleteActionView(MultipleObjectConfirmActionView):
    error_message = _('Error deleting tag "%(instance)s"; %(exception)s')
    model = Tag
    object_permission = permission_tag_delete
    pk_url_kwarg = 'tag_id'
    post_action_redirect = reverse_lazy(viewname='tags:tag_list')
    success_message_single = _('Tag "%(object)s" deleted successfully.')
    success_message_singular = _('%(count)d tag deleted successfully.')
    success_message_plural = _('%(count)d tags deleted successfully.')
    title_single = _('Delete tag: %(object)s.')
    title_singular = _('Delete the %(count)d selected tag.')
    title_plural = _('Delete the %(count)d selected tags.')

    def get_extra_context(self):
        context = {
            'delete_view': True,
            'message': _('Will be removed from all documents.'),
        }

        if self.object_list.count() == 1:
            context.update(
                {
                    'object': self.object_list.first(),
                }
            )

        return context

    def object_action(self, instance, form=None):
        instance.delete()


class TagEditView(SingleObjectEditView):
    fields = ('label', 'color')
    model = Tag
    object_permission = permission_tag_edit
    pk_url_kwarg = 'tag_id'
    post_action_redirect = reverse_lazy(viewname='tags:tag_list')

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _('Edit tag: %s') % self.object,
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class TagListView(SingleObjectListView):
    object_permission = permission_tag_view
    tag_model = Tag

    def get_extra_context(self):
        return {
            'hide_link': True,
            'hide_object': True,
            'no_results_icon': icon_menu_tags,
            'no_results_main_link': link_tag_create.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                'Tags are color coded properties that can be attached or '
                'removed from documents.'
            ),
            'no_results_title': _('No tags available'),
            'title': _('Tags'),
        }

    def get_source_queryset(self):
        queryset = ModelQueryFields.get(model=self.tag_model).get_queryset()
        return queryset.filter(pk__in=self.get_tag_queryset())

    def get_tag_queryset(self):
        return Tag.objects.all()


class TagDocumentListView(ExternalObjectViewMixin, DocumentListView):
    external_object_class = Tag
    external_object_permission = permission_tag_view
    external_object_pk_url_kwarg = 'tag_id'

    def get_document_queryset(self):
        return Document.valid.filter(
            pk__in=self.get_tag().get_documents(
                permission=permission_tag_view, user=self.request.user
            ).values('pk')
        )

    def get_extra_context(self):
        context = super().get_extra_context()
        context.update(
            {
                'object': self.get_tag(),
                'title': _('Documents with the tag: %s') % self.get_tag(),
            }
        )
        return context

    def get_tag(self):
        return self.external_object


class DocumentTagListView(ExternalObjectViewMixin, TagListView):
    external_object_permission = permission_tag_view
    external_object_pk_url_kwarg = 'document_id'
    external_object_queryset = Document.valid
    tag_model = DocumentTag

    def get_extra_context(self):
        context = super().get_extra_context()
        context.update(
            {
                'hide_link': True,
                'no_results_main_link': link_document_tag_multiple_attach.resolve(
                    context=RequestContext(
                        self.request, {'object': self.external_object}
                    )
                ),
                'no_results_title': _('Document has no tags attached'),
                'object': self.external_object,
                'title': _(
                    'Tags for document: %s'
                ) % self.external_object,
            }
        )
        return context

    def get_tag_queryset(self):
        return self.external_object.get_tags(
            permission=permission_tag_view, user=self.request.user
        )


class TagRemoveActionView(MultipleObjectFormActionView):
    form_class = TagMultipleSelectionForm
    object_permission = permission_tag_remove
    pk_url_kwarg = 'document_id'
    source_queryset = Document.valid
    success_message_single = _(
        'Tags removed from document "%(object)s" successfully.'
    )
    success_message_singular = _(
        'Tags removed from %(count)d document successfully.'
    )
    success_message_plural = _(
        'Tags removed from %(count)d documents successfully.'
    )
    title_single = _('Remove tags from document: %(object)s')
    title_singular = _('Remove tags from %(count)d document.')
    title_plural = _('Remove tags from %(count)d documents.')

    def get_extra_context(self):
        context = {
            'submit_icon': icon_document_tag_remove_submit,
            'submit_label': _('Remove'),
        }

        if self.object_list.count() == 1:
            context.update(
                {
                    'object': self.object_list.first(),
                }
            )

        return context

    def get_form_extra_kwargs(self):
        kwargs = {
            'help_text': _('Tags to be removed.'),
            'permission': permission_tag_remove,
            'queryset': Tag.objects.all(),
            'user': self.request.user
        }

        if self.object_list.count() == 1:
            kwargs.update(
                {
                    'queryset': self.object_list.first().tags.all()
                }
            )

        return kwargs

    def get_post_action_redirect(self):
        if self.object_list.count() == 1:
            return reverse(
                viewname='tags:document_tag_list', kwargs={
                    'document_id': self.object_list.first().pk
                }
            )
        else:
            return super().get_post_action_redirect()

    def object_action(self, form, instance):
        for tag in form.cleaned_data['tags']:
            AccessControlList.objects.check_access(
                obj=tag, permissions=(permission_tag_remove,),
                user=self.request.user
            )

            tag._event_actor = self.request.user
            tag.remove_from(document=instance)



import zipfile

try:
    import zlib  # NOQA
    COMPRESSION = zipfile.ZIP_DEFLATED
except ImportError:
    COMPRESSION = zipfile.ZIP_STORED

from django.core.files.base import ContentFile
from django.utils.encoding import force_text

from ..classes import BufferedFile, PassthroughStorage

from .literals import ZIP_CHUNK_SIZE, ZIP_MEMBER_FILENAME


class BufferedZipFile(BufferedFile):
    def __init__(self, *args, **kwargs):
        self.member_name = kwargs.pop('member_name')
        super().__init__(*args, **kwargs)
        self.binary_mode = 'b' in self.mode

        if 'r' in self.mode:
            zip_mode = 'r'
        else:
            zip_mode = 'w'

        self.zip_container_file_object = zipfile.ZipFile(
            file=self.file_object, mode=zip_mode
        )
        self.zip_file_object = self.zip_container_file_object.open(
            name=self.member_name, mode=zip_mode
        )

    def _get_file_object_chunk(self):
        chunk = self.zip_file_object.read(ZIP_CHUNK_SIZE)

        if chunk:
            if self.binary_mode:
                return chunk
            else:
                return force_text(s=chunk)

    def close(self):
        self.zip_file_object.close()
        self.zip_container_file_object.close()
        self.file_object.close()

    def tell(self):
        return self.zip_file_object.tell()

    def write(self, data):
        return self.zip_file_object.write(data)


class ZipCompressedPassthroughStorage(PassthroughStorage):
    def open(self, name, mode='rb', _direct=False):
        next_kwargs = {'name': name}

        if _direct:
            next_kwargs['mode'] = mode

            if issubclass(self.next_storage_class, PassthroughStorage):
                next_kwargs.update({'_direct': _direct})

            return self._call_backend_method(
                method_name='open', kwargs=next_kwargs
            )
        else:
            # Next storage mode is always 'rb+' when reading the zip file
            next_kwargs['mode'] = 'rb+'

            storage_file = self._call_backend_method(
                method_name='open', kwargs=next_kwargs
            )

            return BufferedZipFile(
                file_object=storage_file, member_name=ZIP_MEMBER_FILENAME,
                mode=mode
            )

    def save(self, name, content, max_length=None, _direct=False):
        next_kwargs = {'max_length': max_length, 'name': name}
        if _direct:
            next_kwargs['content'] = content

            if issubclass(self.next_storage_class, PassthroughStorage):
                next_kwargs.update({'_direct': _direct})

            return self._call_backend_method(
                method_name='save', kwargs=next_kwargs
            )
        else:
            if not self._call_backend_method(
                method_name='exists', kwargs={'name': name}
            ):
                name = self._call_backend_method(
                    method_name='save', kwargs={
                        'content': ContentFile(content=''), 'name': name
                    }
                )

            with self._call_backend_method(
                method_name='open', kwargs={
                    'name': name, 'mode': 'wb'
                }
            ) as file_object:
                # From Python: ZipFile requires mode 'r', 'w', 'x', or 'a'
                with zipfile.ZipFile(file=file_object, mode='w', compression=COMPRESSION) as zip_file_object:
                    zip_file_object.writestr(ZIP_MEMBER_FILENAME, content.read())

                    for file in zip_file_object.filelist:
                        file.create_system = 0

            return name



import logging

from furl import furl

from django.contrib import messages
from django.urls import reverse
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _, ungettext
from django.views.generic import RedirectView

from mayan.apps.common.classes import ModelQueryFields
from mayan.apps.common.settings import setting_home_view
from mayan.apps.converter.literals import DEFAULT_ROTATION, DEFAULT_ZOOM_LEVEL
from mayan.apps.views.generics import (
    MultipleObjectConfirmActionView, SimpleView, SingleObjectListView
)
from mayan.apps.views.mixins import ExternalObjectViewMixin
from mayan.apps.views.utils import resolve

from ..forms.document_file_page_forms import DocumentFilePageForm
from ..icons import icon_document_file_page_list
from ..links.document_file_page_links import link_document_file_page_count_update
from ..models.document_file_models import DocumentFile
from ..models.document_file_page_models import DocumentFilePage
from ..permissions import (
    permission_document_file_tools, permission_document_file_view
)
from ..settings import (
    setting_rotation_step, setting_zoom_percent_step, setting_zoom_max_level,
    setting_zoom_min_level
)
from ..tasks import task_document_file_page_count_update

__all__ = (
    'DocumentFilePageListView',
    'DocumentFilePageNavigationFirst', 'DocumentFilePageNavigationLast',
    'DocumentFilePageNavigationNext', 'DocumentFilePageNavigationPrevious',
    'DocumentFilePageView', 'DocumentFilePageViewResetView',
    'DocumentFilePageInteractiveTransformation', 'DocumentFilePageZoomInView',
    'DocumentFilePageZoomOutView', 'DocumentFilePageRotateLeftView',
    'DocumentFilePageRotateRightView'
)
logger = logging.getLogger(name=__name__)


class DocumentFilePageCountUpdateView(MultipleObjectConfirmActionView):
    object_permission = permission_document_file_tools
    pk_url_kwarg = 'document_file_id'
    source_queryset = DocumentFile.valid
    success_message = _(
        '%(count)d document file queued for page count recalculation.'
    )
    success_message_plural = _(
        '%(count)d document files queued for page count recalculation.'
    )

    def get_extra_context(self):
        queryset = self.object_list

        result = {
            'title': ungettext(
                singular='Recalculate the page count of the selected document file?',
                plural='Recalculate the page count of the selected document files?',
                number=queryset.count()
            )
        }

        if queryset.count() == 1:
            result.update(
                {
                    'object': queryset.first(),
                    'title': _(
                        'Recalculate the page count of the document file: %s?'
                    ) % queryset.first()
                }
            )

        return result

    def object_action(self, form, instance):
        task_document_file_page_count_update.apply_async(
            kwargs={'document_file_id': instance.pk}
        )


class DocumentFilePageListView(ExternalObjectViewMixin, SingleObjectListView):
    external_object_permission = permission_document_file_view
    external_object_pk_url_kwarg = 'document_file_id'
    external_object_queryset = DocumentFile.valid

    def get_extra_context(self):
        return {
            'hide_object': True,
            'list_as_items': True,
            'no_results_icon': icon_document_file_page_list,
            'no_results_main_link': link_document_file_page_count_update.resolve(
                request=self.request, resolved_object=self.external_object
            ),
            'no_results_text': _(
                'This could mean that the document file is of a format that '
                'is not supported, that it is corrupted or that the upload '
                'process was interrupted. Use the document file page '
                'recalculation action to attempt to introspect the page '
                'count again.'
            ),
            'no_results_title': _('No document file pages available'),
            'object': self.external_object,
            'title': _('Pages of document file: %s') % self.external_object,
        }

    def get_source_queryset(self):
        queryset = ModelQueryFields.get(model=DocumentFilePage).get_queryset()
        return queryset.filter(pk__in=self.external_object.pages.all())


class DocumentFilePageNavigationBase(ExternalObjectViewMixin, RedirectView):
    external_object_permission = permission_document_file_view
    external_object_pk_url_kwarg = 'document_file_page_id'
    external_object_queryset = DocumentFilePage.valid

    def get_redirect_url(self, *args, **kwargs):
        """
        Attempt to jump to the same kind of view but resolved to a new
        object of the same kind.
        """
        previous_url = self.request.META.get('HTTP_REFERER', None)

        if not previous_url:
            try:
                previous_url = self.external_object.get_absolute_url()
            except AttributeError:
                previous_url = reverse(viewname=setting_home_view.value)

        parsed_url = furl(url=previous_url)

        # Obtain the view name to be able to resolve it back with new keyword
        # arguments.
        resolver_match = resolve(path=force_text(s=parsed_url.path))

        new_kwargs = self.get_new_kwargs()

        if set(new_kwargs) == set(resolver_match.kwargs):
            # It is the same type of object, reuse the URL to stay in the
            # same kind of view but pointing to a new object
            url = reverse(
                viewname=resolver_match.view_name, kwargs=new_kwargs
            )
        else:
            url = parsed_url.path

        # Update just the path to retain the querystring in case there is
        # transformation data.
        parsed_url.path = url

        return parsed_url.tostr()


class DocumentFilePageNavigationFirst(DocumentFilePageNavigationBase):
    def get_new_kwargs(self):
        return {
            'document_file_page_id': self.external_object.siblings.first().pk
        }


class DocumentFilePageNavigationLast(DocumentFilePageNavigationBase):
    def get_new_kwargs(self):
        return {
            'document_file_page_id': self.external_object.siblings.last().pk
        }


class DocumentFilePageNavigationNext(DocumentFilePageNavigationBase):
    def get_new_kwargs(self):
        new_document_file_page = self.external_object.siblings.filter(
            page_number__gt=self.external_object.page_number
        ).first()
        if new_document_file_page:
            return {'document_file_page_id': new_document_file_page.pk}
        else:
            messages.warning(
                message=_(
                    'There are no more pages in this document'
                ), request=self.request
            )
            return {'document_file_page_id': self.external_object.pk}


class DocumentFilePageNavigationPrevious(DocumentFilePageNavigationBase):
    def get_new_kwargs(self):
        new_document_file_page = self.external_object.siblings.filter(
            page_number__lt=self.external_object.page_number
        ).last()
        if new_document_file_page:
            return {'document_file_page_id': new_document_file_page.pk}
        else:
            messages.warning(
                message=_(
                    'You are already at the first page of this document'
                ), request=self.request
            )
            return {'document_file_page_id': self.external_object.pk}


class DocumentFilePageView(ExternalObjectViewMixin, SimpleView):
    external_object_permission = permission_document_file_view
    external_object_pk_url_kwarg = 'document_file_page_id'
    external_object_queryset = DocumentFilePage.valid
    template_name = 'appearance/generic_form.html'

    def get_extra_context(self):
        zoom = int(self.request.GET.get('zoom', DEFAULT_ZOOM_LEVEL))
        rotation = int(self.request.GET.get('rotation', DEFAULT_ROTATION))

        document_file_page_form = DocumentFilePageForm(
            instance=self.external_object, rotation=rotation, zoom=zoom
        )

        base_title = _('Image of: %s') % self.external_object

        if zoom != DEFAULT_ZOOM_LEVEL:
            zoom_text = '({}%)'.format(zoom)
        else:
            zoom_text = ''

        return {
            'form': document_file_page_form,
            'hide_labels': True,
            'object': self.external_object,
            'rotation': rotation,
            'title': ' '.join((base_title, zoom_text)),
            'read_only': True,
            'zoom': zoom,
        }


class DocumentFilePageViewResetView(RedirectView):
    pattern_name = 'documents:document_file_page_view'


class DocumentFilePageInteractiveTransformation(
    ExternalObjectViewMixin, RedirectView
):
    external_object_permission = permission_document_file_view
    external_object_pk_url_kwarg = 'document_file_page_id'
    external_object_queryset = DocumentFilePage.valid

    def get_object(self):
        return self.external_object

    def get_redirect_url(self, *args, **kwargs):
        query_dict = {
            'rotation': self.request.GET.get('rotation', DEFAULT_ROTATION),
            'zoom': self.request.GET.get('zoom', DEFAULT_ZOOM_LEVEL)
        }

        url = furl(
            args=query_dict, path=reverse(
                viewname='documents:document_file_page_view', kwargs={
                    'document_file_page_id': self.external_object.pk
                }
            )

        )

        self.transformation_function(query_dict=query_dict)
        # Refresh query_dict to args reference
        url.args = query_dict

        return url.tostr()


class DocumentFilePageZoomInView(DocumentFilePageInteractiveTransformation):
    def transformation_function(self, query_dict):
        zoom = int(query_dict['zoom']) + setting_zoom_percent_step.value

        if zoom > setting_zoom_max_level.value:
            zoom = setting_zoom_max_level.value

        query_dict['zoom'] = zoom


class DocumentFilePageZoomOutView(DocumentFilePageInteractiveTransformation):
    def transformation_function(self, query_dict):
        zoom = int(query_dict['zoom']) - setting_zoom_percent_step.value

        if zoom < setting_zoom_min_level.value:
            zoom = setting_zoom_min_level.value

        query_dict['zoom'] = zoom


class DocumentFilePageRotateLeftView(DocumentFilePageInteractiveTransformation):
    def transformation_function(self, query_dict):
        query_dict['rotation'] = (
            int(query_dict['rotation']) - setting_rotation_step.value
        ) % 360


class DocumentFilePageRotateRightView(DocumentFilePageInteractiveTransformation):
    def transformation_function(self, query_dict):
        query_dict['rotation'] = (
            int(query_dict['rotation']) + setting_rotation_step.value
        ) % 360



from .classes import Worker

worker_a = Worker(name='worker_a', nice_level=0)
worker_b = Worker(name='worker_b', nice_level=2)
worker_c = Worker(name='worker_c', nice_level=15)
worker_d = Worker(concurrency=1, name='worker_d', nice_level=18)



import json
import logging

from django.db import models, transaction
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _

from django_celery_beat.models import PeriodicTask, IntervalSchedule
from model_utils.managers import InheritanceManager

from mayan.apps.converter.layers import layer_saved_transformations
from mayan.apps.documents.models import DocumentType
from mayan.apps.storage.compressed_files import Archive
from mayan.apps.storage.exceptions import NoMIMETypeMatch

from ..classes import DocumentCreateWizardStep
from ..literals import (
    DEFAULT_INTERVAL, SOURCE_CHOICES, SOURCE_UNCOMPRESS_CHOICES
)

logger = logging.getLogger(name=__name__)


class Source(models.Model):
    label = models.CharField(
        db_index=True, help_text=_('A short text to describe this source.'),
        max_length=128, unique=True, verbose_name=_('Label')
    )
    enabled = models.BooleanField(default=True, verbose_name=_('Enabled'))

    objects = InheritanceManager()

    class Meta:
        ordering = ('label',)
        verbose_name = _('Source')
        verbose_name_plural = _('Sources')

    def __str__(self):
        return '%s' % self.label

    @classmethod
    def class_fullname(cls):
        return force_text(s=dict(SOURCE_CHOICES).get(cls.source_type))

    def clean_up_upload_file(self, upload_file_object):
        """
        Optional method. Non fatal if not defined.
        """

    def fullname(self):
        return ' '.join([self.class_fullname(), '"%s"' % self.label])

    def handle_upload(
        self, file_object, description=None, document_type=None, expand=False,
        label=None, language=None, user=None
    ):
        """
        Handle an upload request from a file object which may be an individual
        document or a compressed file containing multiple documents.
        """
        documents = []
        if not document_type:
            document_type = self.document_type

        kwargs = {
            'description': description, 'document_type': document_type,
            'label': label, 'language': language, 'user': user
        }

        if expand:
            try:
                compressed_file = Archive.open(file_object=file_object)
                for compressed_file_child in compressed_file.members():
                    with compressed_file.open_member(filename=compressed_file_child) as file_object:
                        kwargs.update(
                            {'label': force_text(s=compressed_file_child)}
                        )
                        documents.append(
                            self.upload_document(
                                file_object=file_object, **kwargs
                            )
                        )
            except NoMIMETypeMatch:
                logger.debug(msg='Exception: NoMIMETypeMatch')
                documents.append(
                    self.upload_document(file_object=file_object, **kwargs)
                )
        else:
            documents.append(
                self.upload_document(file_object=file_object, **kwargs)
            )

        # Return a list of newly created documents. Used by the email source
        # to assign the from and subject metadata values.
        return documents

    def get_upload_file_object(self, form_data):
        """
        Optional method. Non fatal if not defined.
        """

    def upload_document(
        self, file_object, document_type, description=None, label=None,
        language=None, querystring=None, user=None
    ):
        """
        Upload an individual document.
        """
        document = None
        try:
            document, document_file = document_type.new_document(
                file_object=file_object, label=label,
                description=description, language=language,
                _user=user
            )
        except Exception as exception:
            logger.critical(
                'Unexpected exception while trying to create new document '
                '"%s" from source "%s"; %s',
                label or file_object.name, self, exception
            )
            if document:
                document.delete(to_trash=False)
            raise
        else:
            if user:
                document.add_as_recent_document_for_user(user=user)

            layer_saved_transformations.copy_transformations(
                source=self, targets=document_file.pages.all()
            )
            DocumentCreateWizardStep.post_upload_process(
                document=document, querystring=querystring
            )
            return document


class InteractiveSource(Source):
    objects = InheritanceManager()

    class Meta:
        verbose_name = _('Interactive source')
        verbose_name_plural = _('Interactive sources')


class OutOfProcessSource(Source):
    is_interactive = False

    objects = models.Manager()

    class Meta:
        verbose_name = _('Out of process')
        verbose_name_plural = _('Out of process')


class IntervalBaseModel(OutOfProcessSource):
    interval = models.PositiveIntegerField(
        default=DEFAULT_INTERVAL,
        help_text=_('Interval in seconds between checks for new documents.'),
        verbose_name=_('Interval')
    )
    document_type = models.ForeignKey(
        help_text=_(
            'Assign a document type to documents uploaded from this source.'
        ), on_delete=models.CASCADE, to=DocumentType,
        related_name='interval_sources', verbose_name=_('Document type')
    )
    uncompress = models.CharField(
        choices=SOURCE_UNCOMPRESS_CHOICES,
        help_text=_('Whether to expand or not, compressed archives.'),
        max_length=1, verbose_name=_('Uncompress')
    )

    objects = models.Manager()

    class Meta:
        verbose_name = _('Interval source')
        verbose_name_plural = _('Interval sources')

    def _delete_periodic_task(self, pk=None):
        try:
            periodic_task = PeriodicTask.objects.get(
                name=self._get_periodic_task_name(pk=pk)
            )

            interval_instance = periodic_task.interval

            if interval_instance.periodictask_set.count() == 1:
                # Only delete the interval if nobody else is using it.
                interval_instance.delete()
            else:
                periodic_task.delete()
        except PeriodicTask.DoesNotExist:
            logger.warning(
                'Tried to delete non existant periodic task "%s"',
                self._get_periodic_task_name(pk)
            )

    def _get_periodic_task_name(self, pk=None):
        return 'check_interval_source-%i' % (pk or self.pk)

    def check_source(self, test=False):
        try:
            self._check_source(test=test)
        except Exception as exception:
            self.error_log.create(
                text='{}; {}'.format(
                    exception.__class__.__name__, exception
                )
            )
            raise
        else:
            self.error_log.all().delete()

    def delete(self, *args, **kwargs):
        pk = self.pk
        with transaction.atomic():
            super().delete(*args, **kwargs)
            self._delete_periodic_task(pk=pk)

    def save(self, *args, **kwargs):
        new_source = not self.pk
        with transaction.atomic():
            super().save(*args, **kwargs)

            if not new_source:
                self._delete_periodic_task()

            interval_instance, created = IntervalSchedule.objects.get_or_create(
                every=self.interval, period='seconds'
            )
            # Create a new interval or reuse someone else's.
            PeriodicTask.objects.create(
                name=self._get_periodic_task_name(),
                interval=interval_instance,
                task='mayan.apps.sources.tasks.task_check_interval_source',
                kwargs=json.dumps(obj={'source_id': self.pk})
            )



from rest_framework import status

from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..events import (
    event_document_comment_created, event_document_comment_deleted,
    event_document_comment_edited
)
from ..models import Comment
from ..permissions import (
    permission_document_comment_create, permission_document_comment_delete,
    permission_document_comment_edit, permission_document_comment_view
)

from .mixins import CommentAPIViewTestMixin, DocumentCommentTestMixin


class CommentAPIViewTestCase(
    CommentAPIViewTestMixin, DocumentCommentTestMixin, DocumentTestMixin,
    BaseAPITestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_user()
        self._create_test_document_stub()

    def test_comment_create_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_comment_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(Comment.objects.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_comment_create_api_view_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_comment_create
        )

        self._clear_events()

        response = self._request_test_comment_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        comment = Comment.objects.first()
        self.assertEqual(Comment.objects.count(), 1)
        self.assertEqual(response.data['id'], comment.pk)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_document_comment)
        self.assertEqual(events[0].verb, event_document_comment_created.id)

    def test_comment_delete_api_view_no_permission(self):
        self._create_test_comment()

        self._clear_events()

        response = self._request_test_comment_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertTrue(self.test_document_comment in Comment.objects.all())

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_comment_delete_api_view_with_access(self):
        self._create_test_comment()
        self.grant_access(
            obj=self.test_document, permission=permission_document_comment_delete
        )

        self._clear_events()

        response = self._request_test_comment_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertFalse(self.test_document_comment in Comment.objects.all())

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_document)
        self.assertEqual(events[0].verb, event_document_comment_deleted.id)

    def test_comment_detail_api_view_no_permission(self):
        self._create_test_comment()

        self._clear_events()

        response = self._request_test_comment_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_comment_detail_api_view_with_access(self):
        self._create_test_comment()
        self.grant_access(
            obj=self.test_document, permission=permission_document_comment_view
        )

        self._clear_events()

        response = self._request_test_comment_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.data['text'], self.test_document_comment.text)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_comment_edit_via_patch_api_view_no_permission(self):
        self._create_test_comment()
        comment_text = self.test_document_comment.text

        self._clear_events()

        response = self._request_test_comment_edit_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.test_document_comment.refresh_from_db()
        self.assertEqual(self.test_document_comment.text, comment_text)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_comment_edit_via_patch_api_view_with_access(self):
        self._create_test_comment()
        self.grant_access(
            obj=self.test_document, permission=permission_document_comment_edit
        )
        comment_text = self.test_document_comment.text

        self._clear_events()

        response = self._request_test_comment_edit_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.test_document_comment.refresh_from_db()
        self.assertNotEqual(self.test_document_comment.text, comment_text)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_document_comment)
        self.assertEqual(events[0].verb, event_document_comment_edited.id)

    def test_comment_list_api_view_no_permission(self):
        self._create_test_comment()

        self._clear_events()

        response = self._request_test_comment_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_comment_list_api_view_with_access(self):
        self._create_test_comment()
        self.grant_access(
            obj=self.test_document, permission=permission_document_comment_view
        )

        self._clear_events()

        response = self._request_test_comment_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['results'][0]['text'], self.test_document_comment.text
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from mayan.apps.appearance.classes import Icon

icon_document_file_all_signature_refresh = Icon(
    driver_name='fontawesome-dual', primary_symbol='file-signature',
    secondary_symbol='recycle'
)
icon_document_file_all_signature_verify = Icon(
    driver_name='fontawesome', symbol='file-signature'
)
icon_document_file_signature_delete = Icon(
    driver_name='fontawesome', symbol='times'
)
icon_document_file_signature_details = Icon(
    driver_name='fontawesome', symbol='file-signature'
)
icon_document_file_signature_detached_create = Icon(
    driver_name='fontawesome', symbol='file-signature'
)
icon_document_file_signature_download = Icon(
    driver_name='fontawesome', symbol='download'
)
icon_document_file_signature_embedded_create = Icon(
    driver_name='fontawesome', symbol='file-signature'
)
icon_document_file_signature_list = Icon(
    driver_name='fontawesome', symbol='file-signature'
)
icon_document_file_signature_upload = Icon(
    driver_name='fontawesome', symbol='upload'
)



DEFAULT_OCR_AUTO_OCR = True
DEFAULT_OCR_BACKEND = 'mayan.apps.ocr.backends.tesseract.Tesseract'
DEFAULT_OCR_BACKEND_ARGUMENTS = {'environment': {'OMP_THREAD_LIMIT': '1'}}

TASK_DOCUMENT_VERSION_PAGE_OCR_RETRY_DELAY = 10
TASK_DOCUMENT_VERSION_PAGE_OCR_TIMEOUT = 10 * 60  # 10 Minutes per page



from django.apps import apps
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType

from ..classes import ErrorLog

from .literals import TEST_ERROR_LOG_ENTRY_RESULT


class LoggingTextMixin:
    def _create_error_log_test_object(self):
        self._create_test_user()
        self.test_model = get_user_model()
        self.test_object = self.test_user

    def _create_error_log_entry(self):
        app_config = apps.get_app_config(app_label='logging')
        self.error_log = ErrorLog(app_config=app_config)
        self.error_log.register_model(
            model=self.test_model, register_permission=True
        )

        self.error_log_entry = self.test_object.error_log.create(
            text=TEST_ERROR_LOG_ENTRY_RESULT
        )


class LoggingViewTestMixin:
    def _request_object_error_log_list_view(self):
        content_type = ContentType.objects.get_for_model(
            model=self.test_object
        )

        return self.get(
            viewname='logging:object_error_list', kwargs={
                'app_label': content_type.app_label,
                'model_name': content_type.model,
                'object_id': self.test_object.pk
            }
        )

    def _request_object_error_log_list_clear_view(self):
        content_type = ContentType.objects.get_for_model(
            model=self.test_object
        )

        return self.post(
            viewname='logging:object_error_list_clear', kwargs={
                'app_label': content_type.app_label,
                'model_name': content_type.model,
                'object_id': self.test_object.pk
            }
        )



import copy
from io import BytesIO
import logging
import os
import shutil

import PIL
from PIL import Image
import sh

from django.apps import apps
from django.contrib.staticfiles.storage import staticfiles_storage
from django.core.exceptions import ImproperlyConfigured
from django.db import transaction
from django.utils.encoding import force_text
from django.utils.functional import cached_property
from django.utils.module_loading import import_string
from django.utils.translation import ugettext_lazy as _

from mayan.apps.appearance.classes import Icon
from mayan.apps.mimetype.api import get_mimetype
from mayan.apps.navigation.classes import Link
from mayan.apps.storage.compressed_files import MsgArchive
from mayan.apps.storage.literals import MSG_MIME_TYPES
from mayan.apps.storage.settings import setting_temporary_directory
from mayan.apps.storage.utils import (
    NamedTemporaryFile, fs_cleanup, mkdtemp
)

from .exceptions import (
    InvalidOfficeFormat, LayerError, OfficeConversionError
)
from .literals import (
    CONVERTER_OFFICE_FILE_MIMETYPES, DEFAULT_LIBREOFFICE_PATH,
    DEFAULT_PAGE_NUMBER, DEFAULT_PILLOW_FORMAT
)
from .settings import (
    setting_graphics_backend, setting_graphics_backend_arguments
)

logger = logging.getLogger(name=__name__)


libreoffice_path = setting_graphics_backend_arguments.value.get(
    'libreoffice_path', DEFAULT_LIBREOFFICE_PATH
)

logger = logging.getLogger(name=__name__)


class AppImageErrorImage:
    _registry = {}

    @classmethod
    def get(cls, name):
        return cls._registry[name]

    def __init__(self, name, image_path=None, template_name=None):
        if name in self.__class__._registry:
            raise ImproperlyConfigured(
                '{} already has a entry named `{}`'.format(__class__, name)
            )

        self.name = name
        self.image_path = image_path
        self.template_name = template_name

        self.__class__._registry[name] = self

    def open(self):
        return staticfiles_storage.open(name=self.image_path, mode='rb')


class ConverterBase:
    @staticmethod
    def get_converter_class():
        return import_string(dotted_path=setting_graphics_backend.value)

    def __init__(self, file_object, mime_type=None):
        self.file_object = file_object
        self.image = None
        self.mime_type = mime_type or get_mimetype(
            file_object=file_object, mimetype_only=False
        )[0]
        self.soffice_file = None
        Image.init()
        try:
            self.command_libreoffice = sh.Command(path=libreoffice_path).bake(
                '--headless', '--convert-to', 'pdf:writer_pdf_Export'
            )
        except sh.CommandNotFound:
            self.command_libreoffice = None

    def convert(self, page_number=DEFAULT_PAGE_NUMBER):
        self.page_number = page_number

    def get_page(self, output_format=None):
        output_format = output_format or setting_graphics_backend_arguments.value.get(
            'pillow_format', DEFAULT_PILLOW_FORMAT
        )

        if not self.image:
            self.seek_page(page_number=0)

        image_buffer = BytesIO()
        new_mode = self.image.mode

        if output_format.upper() == 'JPEG':
            # JPEG doesn't support transparency channel, convert the image to
            # RGB. Removes modes: P and RGBA
            new_mode = 'RGB'

        self.image.convert(new_mode).save(image_buffer, format=output_format)

        image_buffer.seek(0)

        return image_buffer

    def get_page_count(self):
        try:
            self.soffice_file = self.to_pdf()
        except InvalidOfficeFormat as exception:
            logger.debug('Is not an office format document; %s', exception)

    def seek_page(self, page_number):
        """
        Seek the specified page number from the source file object.
        If the file is a paged image get the page if not convert it to a
        paged image format and return the specified page as an image.
        """
        # Starting with #0
        self.file_object.seek(0)

        try:
            self.image = Image.open(fp=self.file_object)
        except IOError:
            # Cannot identify image file
            self.image = self.convert(page_number=page_number)
        except PIL.Image.DecompressionBombError as exception:
            logger.error(
                'Unable to seek document page. Increase the value of '
                'the argument "pillow_maximum_image_pixels" in the '
                'CONVERTER_GRAPHICS_BACKEND_ARGUMENTS setting; %s',
                exception
            )
            raise
        else:
            self.image.seek(page_number)
            self.image.load()

    def soffice(self):
        """
        Executes LibreOffice as a sub process
        """
        if not self.command_libreoffice:
            raise OfficeConversionError(
                _('LibreOffice not installed or not found.')
            )

        with NamedTemporaryFile() as temporary_file_object:
            # Copy the source file object of the converter instance to a
            # named temporary file to be able to pass it to the LibreOffice
            # execution.
            self.file_object.seek(0)
            shutil.copyfileobj(
                fsrc=self.file_object, fdst=temporary_file_object
            )
            self.file_object.seek(0)
            temporary_file_object.seek(0)

            libreoffice_home_directory = mkdtemp()
            args = (
                temporary_file_object.name, '--outdir', setting_temporary_directory.value,
                '-env:UserInstallation=file://{}'.format(
                    os.path.join(
                        libreoffice_home_directory, 'LibreOffice_Conversion'
                    )
                ),
            )

            kwargs = {'_env': {'HOME': libreoffice_home_directory}}

            if self.mime_type == 'text/plain':
                kwargs.update(
                    {'infilter': 'Text (encoded):UTF8,LF,,,'}
                )

            try:
                self.command_libreoffice(*args, **kwargs)
            except sh.ErrorReturnCode as exception:
                temporary_file_object.close()
                raise OfficeConversionError(exception)
            except Exception as exception:
                temporary_file_object.close()
                logger.error(
                    'Exception launching LibreOffice; %s', exception,
                    exc_info=True
                )
                raise
            finally:
                fs_cleanup(filename=libreoffice_home_directory)

            # LibreOffice return a PDF file with the same name as the input
            # provided but with the .pdf extension.

            # Get the converted output file path out of the temporary file
            # name plus the temporary directory

            filename, extension = os.path.splitext(
                os.path.basename(temporary_file_object.name)
            )

            logger.debug('filename: %s', filename)
            logger.debug('extension: %s', extension)

            converted_file_path = os.path.join(
                setting_temporary_directory.value, os.path.extsep.join(
                    (filename, 'pdf')
                )
            )
            logger.debug('converted_file_path: %s', converted_file_path)

        # Don't use context manager with the NamedTemporaryFile on purpose
        # so that it is deleted when the caller closes the file and not
        # before.

        temporary_converted_file_object = NamedTemporaryFile()

        # Copy the LibreOffice output file to a new named temporary file
        # and delete the converted file
        with open(file=converted_file_path, mode='rb') as converted_file_object:
            shutil.copyfileobj(
                fsrc=converted_file_object, fdst=temporary_converted_file_object
            )
        fs_cleanup(filename=converted_file_path)
        temporary_converted_file_object.seek(0)
        return temporary_converted_file_object

    def to_pdf(self):
        # Handle .msg files
        if self.mime_type in MSG_MIME_TYPES:
            archive = MsgArchive.open(file_object=self.file_object)
            members = archive.members()
            if len(members):
                if 'message.txt' in members:
                    self.file_object = archive.open_member(
                        filename='message.txt'
                    )
                else:
                    self.file_object = archive.open_member(
                        filename=members[0]
                    )

                self.mime_type = get_mimetype(
                    file_object=self.file_object, mimetype_only=True
                )[0]

        if self.mime_type in CONVERTER_OFFICE_FILE_MIMETYPES:
            return self.soffice()
        else:
            raise InvalidOfficeFormat(_('Not an office file format.'))

    def transform(self, transformation):
        if not self.image:
            self.seek_page(page_number=0)

        self.image = transformation.execute_on(image=self.image)

    def transform_many(self, transformations):
        if not self.image:
            self.seek_page(page_number=0)

        for transformation in transformations:
            self.image = transformation.execute_on(image=self.image)


class Layer:
    _registry = {}

    @classmethod
    def all(cls):
        return cls._registry.values()

    @classmethod
    def get(cls, name):
        return cls._registry[name]

    @classmethod
    def get_by_value(cls, key, value):
        for name, layer in cls._registry.items():
            if getattr(layer, key) == value:
                return layer

    @classmethod
    def invalidate_cache(cls):
        for layer in cls.all():
            layer.__dict__.pop('stored_layer', None)

    @classmethod
    def update(cls):
        for layer in cls.all():
            layer.stored_layer

    def __init__(
        self, label, name, order, permissions, default=False,
        empty_results_text=None, symbol=None,
    ):
        self.default = default
        self.empty_results_text = empty_results_text
        self.label = label
        self.name = name
        self.order = order
        self.permissions = permissions
        self.symbol = symbol

        # Check order
        layer = self.__class__.get_by_value(key='order', value=self.order)

        if layer:
            raise ImproperlyConfigured(
                'Layer "{}" already has order "{}" requested by layer "{}"'.format(
                    layer.name, order, self.name
                )
            )

        # Check default
        if default:
            layer = self.__class__.get_by_value(key='default', value=True)
            if layer:
                raise ImproperlyConfigured(
                    'Layer "{}" is already the default layer; "{}"'.format(
                        layer.name, self.name
                    )
                )

        self.__class__._registry[name] = self

    def __str__(self):
        return force_text(s=self.label)

    def add_transformation_to(
        self, obj, transformation_class, arguments=None, order=None
    ):
        ContentType = apps.get_model(
            app_label='contenttypes', model_name='ContentType'
        )
        content_type = ContentType.objects.get_for_model(model=obj)
        object_layer, created = self.stored_layer.object_layers.get_or_create(
            content_type=content_type, object_id=obj.pk
        )

        if self in transformation_class._layer_transformations:
            return object_layer.transformations.create(
                arguments=arguments, order=order,
                name=transformation_class.name
            )
        else:
            raise LayerError(
                'Transformation `{}` not registered for layer `{}`.'.format(
                    transformation_class, self
                )
            )

    def copy_transformations(self, source, targets, delete_existing=False):
        """
        Copy transformation from source to all targets
        """
        ContentType = apps.get_model(
            app_label='contenttypes', model_name='ContentType'
        )

        transformations = self.get_transformations_for(obj=source)

        with transaction.atomic():
            for target in targets:
                content_type = ContentType.objects.get_for_model(model=target)
                object_layer, created = self.stored_layer.object_layers.get_or_create(
                    content_type=content_type, object_id=target.pk
                )
                if delete_existing:
                    object_layer.transformations.all().delete()

                for transformation in transformations:
                    object_layer.transformations.create(
                        order=transformation.order,
                        name=transformation.name,
                        arguments=transformation.arguments,
                    )

    def get_empty_results_text(self):
        if self.empty_results_text:
            return self.empty_results_text
        else:
            return _(
                'Transformations allow changing the visual appearance '
                'of documents without making permanent changes to the '
                'document file themselves.'
            )

    def get_icon(self):
        return Icon(driver_name='fontawesome', symbol=self.symbol)

    def get_model_instance(self):
        StoredLayer = apps.get_model(
            app_label='converter', model_name='StoredLayer'
        )
        stored_layer, created = StoredLayer.objects.update_or_create(
            name=self.name, defaults={'order': self.order}
        )

        return stored_layer

    def get_permission(self, action):
        return self.permissions.get(action, None)

    def get_transformations_for(self, obj, as_classes=False):
        """
        as_classes == True returns the transformation classes from .classes
        ready to be feed to the converter class
        """
        LayerTransformation = apps.get_model(
            app_label='converter', model_name='LayerTransformation'
        )

        return LayerTransformation.objects.get_for_object(
            obj=obj, as_classes=as_classes,
            only_stored_layer=self.stored_layer
        )

    @cached_property
    def stored_layer(self):
        return self.get_model_instance()


class LayerLink(Link):
    @staticmethod
    def set_icon(instance, layer):
        if instance.action == 'list':
            if layer.symbol:
                instance.icon = layer.get_icon()

    def __init__(self, action, layer, object_name=None, **kwargs):
        super().__init__(**kwargs)
        self.action = action
        self.layer = layer
        self.object_name = object_name or _('transformation')

        permission = layer.get_permission(action=action)
        if permission:
            self.permissions = (permission,)

        if action == 'list':
            self.kwargs = LayerLinkKwargsFactory(
                layer_name=layer.name
            ).get_kwargs_function()

        if action in ('create', 'select'):
            self.kwargs = LayerLinkKwargsFactory().get_kwargs_function()

        LayerLink.set_icon(instance=self, layer=layer)

    def copy(self, layer):
        result = copy.copy(self)
        result.kwargs = LayerLinkKwargsFactory(
            layer_name=layer.name
        ).get_kwargs_function()
        result._layer_name = layer.name

        LayerLink.set_icon(instance=result, layer=layer)

        return result

    @cached_property
    def layer_name(self):
        return getattr(
            self, '_layer_name', Layer.get_by_value(
                key='default', value=True
            ).name
        )


class LayerLinkKwargsFactory:
    def __init__(self, layer_name=None, variable_name='resolved_object'):
        self.layer_name = layer_name
        self.variable_name = variable_name

    def get_kwargs_function(self):
        def get_kwargs(context):
            ContentType = apps.get_model(
                app_label='contenttypes', model_name='ContentType'
            )

            content_type = ContentType.objects.get_for_model(
                context[self.variable_name]
            )
            default_layer = Layer.get_by_value(key='default', value=True)
            return {
                'app_label': '"{}"'.format(content_type.app_label),
                'model_name': '"{}"'.format(content_type.model),
                'object_id': '{}.pk'.format(self.variable_name),
                'layer_name': '"{}"'.format(
                    self.layer_name or context.get(
                        'layer_name', default_layer.name
                    )
                )
            }

        return get_kwargs



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('document_parsing', '0006_rename_documentversionparseerror_model'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='documentfileparseerror',
            options={'ordering': ('datetime_submitted',), 'verbose_name': 'Document file parse error', 'verbose_name_plural': 'Document file parse errors'},
        ),
        migrations.AlterField(
            model_name='documentfileparseerror',
            name='document_file',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='parsing_errors', to='documents.DocumentFile', verbose_name='Document file'),
        ),
    ]



__title__ = 'Mayan EDMS'
__version__ = '4.0.3'
__build__ = 0x040003
__build_string__ = 'v4.0.3_Thu Jun 3 06:00:42 2021 -0400'
__django_version__ = '2.2'
__author__ = 'Roberto Rosario'
__author_email__ = 'roberto.rosario@mayan-edms.com'
__description__ = 'Free Open Source Electronic Document Management System'
__license__ = 'Apache 2.0'
__copyright_short__ = '2011 Roberto Rosario'
__copyright__ = '{} {}'.format('Copyright', __copyright_short__)
__website__ = 'https://www.mayan-edms.com'



from django.utils.translation import ugettext_lazy as _

from mayan.apps.navigation.classes import Link

from .icons import (
    icon_document_send, icon_document_send_link, icon_user_mailer_create,
    icon_user_mailer_delete, icon_user_mailer_edit, icon_user_mailer_list,
    icon_user_mailer_setup, icon_user_mailer_test
)
from .permissions import (
    permission_mailing_link, permission_mailing_send_document,
    permission_user_mailer_create, permission_user_mailer_delete,
    permission_user_mailer_edit, permission_user_mailer_use,
    permission_user_mailer_view
)

link_send_document = Link(
    args='resolved_object.pk', icon=icon_document_send,
    permissions=(permission_mailing_send_document,),
    text=_('Email document'), view='mailer:send_document'
)
link_send_document_link = Link(
    args='resolved_object.pk', icon=icon_document_send_link,
    permissions=(permission_mailing_link,),
    text=_('Email link'), view='mailer:send_document_link'
)
link_send_multiple_document = Link(
    icon=icon_document_send, text=_('Email document'),
    view='mailer:send_multiple_document'
)
link_send_multiple_document_link = Link(
    icon=icon_document_send_link, text=_('Email link'),
    view='mailer:send_multiple_document_link'
)
link_user_mailer_create = Link(
    icon=icon_user_mailer_create,
    permissions=(permission_user_mailer_create,),
    text=_('Create mailing profile'),
    view='mailer:user_mailer_backend_selection'
)
link_user_mailer_delete = Link(
    args='resolved_object.pk', icon=icon_user_mailer_delete,
    permissions=(permission_user_mailer_delete,), tags='dangerous',
    text=_('Delete'), view='mailer:user_mailer_delete'
)
link_user_mailer_edit = Link(
    args='object.pk', icon=icon_user_mailer_edit,
    permissions=(permission_user_mailer_edit,), text=_('Edit'),
    view='mailer:user_mailer_edit'
)
link_user_mailer_list = Link(
    icon=icon_user_mailer_list,
    permissions=(permission_user_mailer_view,),
    text=_('Mailing profiles list'), view='mailer:user_mailer_list'
)
link_user_mailer_setup = Link(
    icon=icon_user_mailer_setup,
    permissions=(permission_user_mailer_view,), text=_('Mailing profiles'),
    view='mailer:user_mailer_list'
)
link_user_mailer_test = Link(
    args='object.pk', icon=icon_user_mailer_test,
    permissions=(permission_user_mailer_use,), text=_('Test'),
    view='mailer:user_mailer_test'
)



from django import forms
from django.utils.translation import ugettext_lazy as _

from mayan.apps.views.widgets import TextAreaDiv


class MessageDetailForm(forms.Form):
    body = forms.CharField(
        label=_('Body'),
        widget=TextAreaDiv(
            attrs={
                'class': 'views-text-wrap text_area_div full-height',
                'data-height-difference': 360,
            }
        )
    )

    def __init__(self, *args, **kwargs):
        self.instance = kwargs.pop('instance', None)
        super().__init__(*args, **kwargs)
        self.fields['body'].initial = self.instance.body



from mayan.apps.rest_api import generics

from ..models.document_models import RecentlyCreatedDocument
from ..permissions import permission_document_view
from ..serializers.document_serializers import DocumentSerializer


class APIRecentlyCreatedDocumentListView(generics.ListAPIView):
    """
    get: Return a list of the recently created documents.
    """
    mayan_object_permissions = {
        'GET': (permission_document_view,)
    }
    queryset = RecentlyCreatedDocument.recently_created.all()
    serializer_class = DocumentSerializer



from mayan.apps.testing.tests.base import GenericViewTestCase
from mayan.apps.user_management.permissions import permission_group_edit
from mayan.apps.user_management.tests.mixins import GroupTestMixin

from ..events import event_role_created, event_role_edited
from ..models import Role
from ..permissions import (
    permission_role_create, permission_role_delete, permission_role_edit,
    permission_role_view
)

from .mixins import (
    GroupRoleAddRemoveViewTestMixin, RoleGroupAddRemoveViewTestMixin,
    RolePermissionAddRemoveViewTestMixin, RoleTestMixin, RoleViewTestMixin
)


class RoleViewsTestCase(
    RoleTestMixin, RoleViewTestMixin, GenericViewTestCase
):
    def test_role_creation_view_no_permission(self):
        role_count = Role.objects.count()

        self._clear_events()

        response = self._request_test_role_create_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(Role.objects.count(), role_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_creation_view_with_permission(self):
        self.grant_permission(permission=permission_role_create)
        role_count = Role.objects.count()

        self._clear_events()

        response = self._request_test_role_create_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Role.objects.count(), role_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_role)
        self.assertEqual(events[0].verb, event_role_created.id)

    def test_role_delete_view_no_permission(self):
        self._create_test_role()

        role_count = Role.objects.count()

        self._clear_events()

        response = self._request_test_role_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Role.objects.count(), role_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_delete_view_with_access(self):
        self._create_test_role()
        self.grant_access(
            obj=self.test_role, permission=permission_role_delete
        )

        role_count = Role.objects.count()

        self._clear_events()

        response = self._request_test_role_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Role.objects.count(), role_count - 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_edit_view_no_permission(self):
        self._create_test_role()
        role_label = self.test_role.label

        self._clear_events()

        response = self._request_test_role_edit_view()

        self.assertEqual(response.status_code, 404)

        self.test_role.refresh_from_db()
        self.assertEqual(self.test_role.label, role_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_edit_view_with_access(self):
        self._create_test_role()
        self.grant_access(obj=self.test_role, permission=permission_role_edit)

        role_label = self.test_role.label

        self._clear_events()

        response = self._request_test_role_edit_view()
        self.assertEqual(response.status_code, 302)

        self.test_role.refresh_from_db()
        self.assertNotEqual(self.test_role.label, role_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_role)
        self.assertEqual(events[0].verb, event_role_edited.id)

    def test_role_list_view_no_permission(self):
        self._create_test_role()

        self._clear_events()

        response = self._request_test_role_list_view()
        self.assertEqual(response.status_code, 200)
        self.assertNotContains(
            response=response, text=self.test_role.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_list_view_with_access(self):
        self._create_test_role()
        self.grant_access(obj=self.test_role, permission=permission_role_view)

        self._clear_events()

        response = self._request_test_role_list_view()
        self.assertContains(
            response=response, text=self.test_role.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class RoleGroupAddRemoveViewTestCase(
    GroupTestMixin, RoleGroupAddRemoveViewTestMixin, RoleTestMixin,
    GenericViewTestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_role()
        self._create_test_group()

    def test_role_group_add_remove_get_view_no_permission(self):
        self.test_role.groups.add(self.test_group)

        self._clear_events()

        response = self._request_test_role_group_add_remove_get_view()
        self.assertNotContains(
            response=response, text=str(self.test_role),
            status_code=404
        )
        self.assertNotContains(
            response=response, text=str(self.test_group),
            status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_group_add_remove_get_view_with_role_access(self):
        self.test_role.groups.add(self.test_group)

        self.grant_access(
            obj=self.test_role,
            permission=permission_role_edit
        )

        self._clear_events()

        response = self._request_test_role_group_add_remove_get_view()
        self.assertContains(
            response=response, text=str(self.test_role),
            status_code=200
        )
        self.assertNotContains(
            response=response, text=str(self.test_group),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_group_add_remove_get_view_with_group_access(self):
        self.test_role.groups.add(self.test_group)

        self.grant_access(
            obj=self.test_group,
            permission=permission_group_edit
        )

        self._clear_events()

        response = self._request_test_role_group_add_remove_get_view()
        self.assertNotContains(
            response=response, text=str(self.test_role),
            status_code=404
        )
        self.assertNotContains(
            response=response, text=str(self.test_group),
            status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_group_add_remove_get_view_with_full_access(self):
        self.test_role.groups.add(self.test_group)

        self.grant_access(
            obj=self.test_role,
            permission=permission_role_edit
        )
        self.grant_access(
            obj=self.test_group,
            permission=permission_group_edit
        )

        self._clear_events()

        response = self._request_test_role_group_add_remove_get_view()
        self.assertContains(
            response=response, text=str(self.test_role),
            status_code=200
        )
        self.assertContains(
            response=response, text=str(self.test_group),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_group_add_view_no_permission(self):
        self._clear_events()

        response = self._request_test_role_group_add_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self.test_group not in self.test_role.groups.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_group_add_view_with_role_access(self):
        self.grant_access(
            obj=self.test_role,
            permission=permission_role_edit
        )

        self._clear_events()

        response = self._request_test_role_group_add_view()
        self.assertEqual(response.status_code, 200)

        self.assertTrue(
            self.test_group not in self.test_role.groups.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_group_add_view_with_group_access(self):
        self.grant_access(
            obj=self.test_group,
            permission=permission_group_edit
        )

        self._clear_events()

        response = self._request_test_role_group_add_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self.test_group not in self.test_role.groups.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_group_add_view_with_full_access(self):
        self.grant_access(
            obj=self.test_role,
            permission=permission_role_edit
        )
        self.grant_access(
            obj=self.test_group,
            permission=permission_group_edit
        )

        self._clear_events()

        response = self._request_test_role_group_add_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self.test_group in self.test_role.groups.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_group)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_role)
        self.assertEqual(events[0].verb, event_role_edited.id)

    def test_role_group_remove_view_no_permission(self):
        self.test_role.groups.add(self.test_group)

        self._clear_events()

        response = self._request_test_role_group_remove_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self.test_group in self.test_role.groups.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_group_remove_view_with_role_access(self):
        self.test_role.groups.add(self.test_group)

        self.grant_access(
            obj=self.test_role,
            permission=permission_role_edit
        )

        self._clear_events()

        response = self._request_test_role_group_remove_view()
        self.assertEqual(response.status_code, 200)

        self.assertTrue(
            self.test_group in self.test_role.groups.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_group_remove_view_with_group_access(self):
        self.test_role.groups.add(self.test_group)

        self.grant_access(
            obj=self.test_group,
            permission=permission_group_edit
        )

        self._clear_events()

        response = self._request_test_role_group_remove_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self.test_group in self.test_role.groups.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_group_remove_view_with_full_access(self):
        self.test_role.groups.add(self.test_group)

        self.grant_access(
            obj=self.test_role,
            permission=permission_role_edit
        )
        self.grant_access(
            obj=self.test_group,
            permission=permission_group_edit
        )

        self._clear_events()

        response = self._request_test_role_group_remove_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self.test_group not in self.test_role.groups.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_group)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_role)
        self.assertEqual(events[0].verb, event_role_edited.id)


class RolePermissionAddRemoveViewTestCase(
    RolePermissionAddRemoveViewTestMixin, RoleTestMixin, GenericViewTestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_role()
        self._create_test_permission()

    def test_role_permission_add_remove_get_view_no_permission(self):
        self.test_role.permissions.add(self.test_permission.stored_permission)

        self._clear_events()

        response = self._request_test_role_permission_add_remove_get_view()
        self.assertNotContains(
            response=response, text=str(self.test_role),
            status_code=404
        )
        self.assertNotContains(
            response=response, text=str(self.test_permission),
            status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_permission_add_remove_get_view_with_access(self):
        """
        Tests that a defined permission is available before it is ever used
        or referenced. Tests the runtime permission to stored permission
        cache initialization.
        https://forum.mayan-edms.com/viewtopic.php?f=7&t=1614
        GitLab issue #757 "Permissions list does not show an object until
        one has been created"
        """
        self.test_role.permissions.add(self.test_permission.stored_permission)

        self.grant_access(
            obj=self.test_role,
            permission=permission_role_edit
        )

        self._clear_events()

        response = self._request_test_role_permission_add_remove_get_view()
        self.assertContains(
            response=response, text=str(self.test_role),
            status_code=200
        )
        self.assertContains(
            response=response, text=str(self.test_permission),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_permission_add_view_no_permission(self):
        self._clear_events()

        response = self._request_test_role_permission_add_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self.test_permission.stored_permission not in self.test_role.permissions.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_permission_add_view_with_access(self):
        self.grant_access(
            obj=self.test_role,
            permission=permission_role_edit
        )

        self._clear_events()

        response = self._request_test_role_permission_add_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self.test_permission.stored_permission in self.test_role.permissions.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(
            events[0].action_object, self.test_permission.stored_permission
        )
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_role)
        self.assertEqual(events[0].verb, event_role_edited.id)

    def test_role_permission_remove_view_no_permission(self):
        self.test_role.permissions.add(self.test_permission.stored_permission)

        self._clear_events()

        response = self._request_test_role_permission_remove_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self.test_permission.stored_permission in self.test_role.permissions.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_role_permission_remove_view_with_access(self):
        self.test_role.permissions.add(self.test_permission.stored_permission)

        self.grant_access(
            obj=self.test_role,
            permission=permission_role_edit
        )

        self._clear_events()

        response = self._request_test_role_permission_remove_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self.test_permission.stored_permission not in self.test_role.permissions.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(
            events[0].action_object, self.test_permission.stored_permission
        )
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_role)
        self.assertEqual(events[0].verb, event_role_edited.id)


class GroupRoleAddRemoveViewTestCase(
    GroupTestMixin, GroupRoleAddRemoveViewTestMixin, RoleTestMixin,
    GenericViewTestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_group()
        self._create_test_role()

    def test_group_role_add_remove_get_view_no_permission(self):
        self.test_group.roles.add(self.test_role)

        self._clear_events()

        response = self._request_test_group_role_add_remove_get_view()
        self.assertNotContains(
            response=response, text=str(self.test_group),
            status_code=404
        )
        self.assertNotContains(
            response=response, text=str(self.test_role),
            status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_role_add_remove_get_view_with_group_access(self):
        self.test_group.roles.add(self.test_role)

        self.grant_access(
            obj=self.test_group,
            permission=permission_group_edit
        )

        self._clear_events()

        response = self._request_test_group_role_add_remove_get_view()
        self.assertContains(
            response=response, text=str(self.test_group),
            status_code=200
        )
        self.assertNotContains(
            response=response, text=str(self.test_role),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_role_add_remove_get_view_with_role_access(self):
        self.test_group.roles.add(self.test_role)

        self.grant_access(
            obj=self.test_role,
            permission=permission_role_edit
        )

        self._clear_events()

        response = self._request_test_group_role_add_remove_get_view()
        self.assertNotContains(
            response=response, text=str(self.test_group),
            status_code=404
        )
        self.assertNotContains(
            response=response, text=str(self.test_role),
            status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_role_add_remove_get_view_with_full_access(self):
        self.test_group.roles.add(self.test_role)

        self.grant_access(
            obj=self.test_group,
            permission=permission_group_edit
        )
        self.grant_access(
            obj=self.test_role,
            permission=permission_role_edit
        )

        self._clear_events()

        response = self._request_test_group_role_add_remove_get_view()
        self.assertContains(
            response=response, text=str(self.test_group),
            status_code=200
        )
        self.assertContains(
            response=response, text=str(self.test_role),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_role_add_view_no_permission(self):
        self._clear_events()

        response = self._request_test_group_role_add_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self.test_role not in self.test_group.roles.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_role_add_view_with_group_access(self):
        self.grant_access(
            obj=self.test_group,
            permission=permission_group_edit
        )

        self._clear_events()

        response = self._request_test_group_role_add_view()
        self.assertEqual(response.status_code, 200)

        self.assertTrue(
            self.test_role not in self.test_group.roles.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_role_add_view_with_role_access(self):
        self.grant_access(
            obj=self.test_role,
            permission=permission_role_edit
        )

        self._clear_events()

        response = self._request_test_group_role_add_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self.test_role not in self.test_group.roles.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_role_add_view_with_full_access(self):
        self.grant_access(
            obj=self.test_group,
            permission=permission_group_edit
        )
        self.grant_access(
            obj=self.test_role,
            permission=permission_role_edit
        )

        self._clear_events()

        response = self._request_test_group_role_add_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self.test_role in self.test_group.roles.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_group)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_role)
        self.assertEqual(events[0].verb, event_role_edited.id)

    def test_group_role_remove_view_no_permission(self):
        self.test_group.roles.add(self.test_role)

        self._clear_events()

        response = self._request_test_group_role_remove_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self.test_role in self.test_group.roles.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_role_remove_view_with_group_access(self):
        self.test_group.roles.add(self.test_role)

        self.grant_access(
            obj=self.test_group,
            permission=permission_group_edit
        )

        self._clear_events()

        response = self._request_test_group_role_remove_view()
        self.assertEqual(response.status_code, 200)

        self.assertTrue(
            self.test_role in self.test_group.roles.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_role_remove_view_with_role_access(self):
        self.test_group.roles.add(self.test_role)

        self.grant_access(
            obj=self.test_role,
            permission=permission_role_edit
        )

        self._clear_events()

        response = self._request_test_group_role_remove_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self.test_role in self.test_group.roles.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_group_role_remove_view_with_full_access(self):
        self.test_group.roles.add(self.test_role)

        self.grant_access(
            obj=self.test_group,
            permission=permission_group_edit
        )
        self.grant_access(
            obj=self.test_role,
            permission=permission_role_edit
        )

        self._clear_events()

        response = self._request_test_group_role_remove_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self.test_role not in self.test_group.roles.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_group)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_role)
        self.assertEqual(events[0].verb, event_role_edited.id)



from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.testing.tests.base import BaseTestCase

from .mixins import TagTestMixin


class TagDocumentTestCase(DocumentTestMixin, TagTestMixin, BaseTestCase):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()

    def test_document_addition(self):
        self._create_test_tag()

        self.test_tag.attach_to(document=self.test_document)

        self.assertTrue(self.test_document in self.test_tag.documents.all())

    def test_document_remove(self):
        self._create_test_tag(add_test_document=True)

        self.test_tag.remove_from(document=self.test_document)

        self.assertTrue(
            self.test_document not in self.test_tag.documents.all()
        )


class TagModuleTestCase(TagTestMixin, BaseTestCase):
    def test_method_get_absolute_url(self):
        self._create_test_tag()

        self.assertTrue(self.test_tag.get_absolute_url())



import json
import logging
from pathlib import Path

import sh

from django.utils.translation import ugettext_lazy as _

from mayan.apps.storage.utils import fs_cleanup, mkdtemp

from ..literals import DEFAULT_EXIF_PATH
from ..classes import FileMetadataDriver
from ..settings import setting_drivers_arguments

logger = logging.getLogger(name=__name__)


class EXIFToolDriver(FileMetadataDriver):
    label = _('EXIF Tool')
    internal_name = 'exiftool'

    def __init__(self, *args, **kwargs):
        auto_initialize = kwargs.pop('auto_initialize', True)

        self.read_settings()

        if auto_initialize:
            try:
                self.command_exiftool = sh.Command(path=self.exiftool_path)
            except sh.CommandNotFound:
                self.command_exiftool = None
            else:
                self.command_exiftool = self.command_exiftool.bake('-j')

    def _process(self, document_file):
        if self.command_exiftool:
            temporary_folder = mkdtemp()
            path_temporary_file = Path(temporary_folder, document_file.document.label)

            try:
                with path_temporary_file.open(mode='xb') as temporary_fileobject:
                    document_file.save_to_file(file_object=temporary_fileobject)
                    temporary_fileobject.seek(0)
                    try:
                        result = self.command_exiftool(temporary_fileobject.name)
                    except sh.ErrorReturnCode_1 as exception:
                        result = json.loads(s=exception.stdout)[0]
                        if result.get('Error', '') == 'Unknown file type':
                            # Not a fatal error
                            return result
                    else:
                        return json.loads(s=result.stdout)[0]
            except Exception as exception:
                logger.error(
                    'Error processing document file: %s; %s',
                    document_file, exception, exc_info=True
                )
                raise
            finally:
                fs_cleanup(filename=str(path_temporary_file))
        else:
            logger.warning(
                'EXIFTool binary not found, not processing document '
                'file: %s', document_file
            )

    def read_settings(self):
        self.exiftool_path = setting_drivers_arguments.value.get(
            'exif_driver', {}
        ).get('exiftool_path', DEFAULT_EXIF_PATH)


EXIFToolDriver.register(mimetypes=('*',))



from django.contrib import admin

from mayan.apps.common.admin_mixins import ReadOnlyAdminMixin

from .models import ErrorLog, ErrorLogPartition, ErrorLogPartitionEntry


class ErrorLogPartitionEntryInline(admin.TabularInline):
    model = ErrorLogPartitionEntry


class ErrorLogPartitionInline(admin.StackedInline):
    classes = ('collapse-open',)
    model = ErrorLogPartition


@admin.register(ErrorLog)
class ErrorLogAdmin(ReadOnlyAdminMixin, admin.ModelAdmin):
    inlines = (ErrorLogPartitionInline,)
    list_display = ('name', 'app_label')
    readonly_fields = list_display


@admin.register(ErrorLogPartition)
class ErrorLogPartitionAdmin(ReadOnlyAdminMixin, admin.ModelAdmin):
    inlines = (ErrorLogPartitionEntryInline,)
    list_display = ('name', 'error_log', 'get_model_instance')
    readonly_fields = list_display



from django.utils.translation import ugettext_lazy as _

from mayan.apps.task_manager.classes import CeleryQueue
from mayan.apps.task_manager.workers import worker_a

queue_converter = CeleryQueue(
    label=_('Converter'), name='converter', transient=True, worker=worker_a
)

queue_converter.add_task_type(
    dotted_path='mayan.apps.converter.tasks.task_asset_image_generate',
    label=_('Generate a image of an asset.'),
    name='task_asset_image_generate',
)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(label=_('Documents'), name='documents')

# Document

event_document_created = namespace.add_event_type(
    label=_('Document created'), name='document_create'
)
event_document_edited = namespace.add_event_type(
    label=_('Document edited'), name='document_edit'
)
event_document_viewed = namespace.add_event_type(
    label=_('Document viewed'), name='document_view'
)

# Document File

event_document_file_created = namespace.add_event_type(
    label=_('Document file created'), name='document_file_created'
)
event_document_file_deleted = namespace.add_event_type(
    label=_('Document file deleted'), name='document_file_deleted'
)
event_document_file_downloaded = namespace.add_event_type(
    label=_('Document file downloaded'), name='document_file_downloaded'
)
event_document_file_edited = namespace.add_event_type(
    label=_('Document file edited'), name='document_file_edited'
)
event_document_file_new = namespace.add_event_type(
    label=_('Document file uploaded'), name='document_file_new'
)

# Document type

event_document_type_created = namespace.add_event_type(
    label=_('Document type created'), name='document_type_created'
)
event_document_type_edited = namespace.add_event_type(
    label=_('Document type edited'), name='document_type_edit'
)
event_document_type_quick_label_created = namespace.add_event_type(
    label=_('Document type quick label created'),
    name='document_type_quick_label_created'
)
event_document_type_quick_label_deleted = namespace.add_event_type(
    label=_('Document type quick label deleted'),
    name='document_type_quick_label_deleted'
)
event_document_type_quick_label_edited = namespace.add_event_type(
    label=_('Document type quick label edited'),
    name='document_type_quick_label_edited'
)
# The type of an existing document is changed to another type
event_document_type_changed = namespace.add_event_type(
    label=_('Document type changed'), name='document_type_change'
)

# Document version

event_document_version_created = namespace.add_event_type(
    label=_('Document version created'), name='document_version_created'
)
event_document_version_deleted = namespace.add_event_type(
    label=_('Document version deleted'), name='document_version_deleted'
)
event_document_version_edited = namespace.add_event_type(
    label=_('Document version edited'), name='document_version_edited'
)
event_document_version_exported = namespace.add_event_type(
    label=_('Document version exported'), name='document_version_exported'
)

# Document version page

event_document_version_page_created = namespace.add_event_type(
    label=_('Document version page created'),
    name='document_version_page_created'
)
event_document_version_page_deleted = namespace.add_event_type(
    label=_('Document version page deleted'),
    name='document_version_page_deleted'
)
event_document_version_page_edited = namespace.add_event_type(
    label=_('Document version page edited'),
    name='document_version_page_edited'
)

# Trashed document

event_document_trashed = namespace.add_event_type(
    label=_('Document trashed'), name='document_trashed'
)
event_trashed_document_deleted = namespace.add_event_type(
    label=_('Trashed document deleted'), name='trashed_document_deleted'
)
event_trashed_document_restored = namespace.add_event_type(
    label=_('Trashed document restored'), name='trashed_document_restored'
)

# Historic events

event_document_download = namespace.add_event_type(
    label=_('Document downloaded (historic)'), name='document_download'
)
event_document_file_pre_save = namespace.add_event_type(
    label=_('Document file created (historic)'),
    name='document_file_pre_save'
)
event_document_version_revert = namespace.add_event_type(
    label=_('Document version reverted (historic)'),
    name='document_version_revert'
)
event_document_version_new = namespace.add_event_type(
    label=_('Document version uploaded (historic)'),
    name='document_new_version'
)



from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.document_indexing.models import (
    IndexInstanceNode, IndexTemplate
)

from .literals import (
    TEST_INDEX_TEMPLATE_LABEL, TEST_INDEX_TEMPLATE_METADATA_EXPRESSION
)
from .mixins.workflow_template_mixins import WorkflowTemplateTestMixin


class DocumentStateIndexingTestCase(
    WorkflowTemplateTestMixin, GenericDocumentTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_workflow_template(add_test_document_type=True)
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_transition()

    def _create_test_index_template(self):
        # Create empty index
        index_template = IndexTemplate.objects.create(
            label=TEST_INDEX_TEMPLATE_LABEL
        )

        # Add our document type to the new index
        index_template.document_types.add(self.test_document_type)

        # Create simple index template
        root = index_template.template_root
        index_template.node_templates.create(
            parent=root, expression=TEST_INDEX_TEMPLATE_METADATA_EXPRESSION,
            link_documents=True
        )

    def test_workflow_indexing_initial_state(self):
        self._create_test_index_template()
        self._upload_test_document()

        self.assertEqual(
            list(
                IndexInstanceNode.objects.values_list('value', flat=True)
            ), ['', str(self.test_workflow_template_states[0])]
        )

    def test_workflow_indexing_transition(self):
        self._create_test_index_template()
        self._create_test_document_stub()

        self.test_document.workflows.first().do_transition(
            transition=self.test_workflow_template_transition
        )

        self.assertEqual(
            list(
                IndexInstanceNode.objects.values_list('value', flat=True)
            ), ['', str(self.test_workflow_template_states[1])]
        )

    def test_workflow_indexing_document_delete(self):
        self._create_test_index_template()
        self._create_test_document_stub()

        self.test_document.workflows.first().do_transition(
            transition=self.test_workflow_template_transition
        )

        self.test_document.delete(to_trash=False)

        self.assertEqual(
            list(
                IndexInstanceNode.objects.values_list('value', flat=True)
            ), ['']
        )



from django.utils.encoding import force_bytes

from mayan.apps.storage.classes import DefinedStorage
from mayan.apps.storage.utils import fs_cleanup, mkdtemp

from ..models import Cache
from ..tasks import task_cache_partition_purge, task_cache_purge

from .literals import (
    TEST_CACHE_MAXIMUM_SIZE, TEST_CACHE_PARTITION_FILE_FILENAME,
    TEST_CACHE_PARTITION_FILE_SIZE, TEST_CACHE_PARTITION_NAME,
    TEST_STORAGE_NAME_FILE_CACHING_TEST_STORAGE
)


class CachePartitionViewTestMixin:
    def _request_test_object_file_cache_partition_purge_view(self):
        return self.post(
            viewname='file_caching:cache_partitions_purge',
            kwargs=self.test_object_view_kwargs
        )


class CacheTestMixin:
    def setUp(self):
        super().setUp()
        self.temporary_directory = mkdtemp()
        DefinedStorage(
            dotted_path='django.core.files.storage.FileSystemStorage',
            label='File caching test storage',
            name=TEST_STORAGE_NAME_FILE_CACHING_TEST_STORAGE,
            kwargs={'location': self.temporary_directory}
        )
        self.test_cache_partition_files = []

    def tearDown(self):
        fs_cleanup(filename=self.temporary_directory)
        super().tearDown()

    def _create_test_cache(self, extra_data=None):
        data = {
            'defined_storage_name': TEST_STORAGE_NAME_FILE_CACHING_TEST_STORAGE,
            'maximum_size': TEST_CACHE_MAXIMUM_SIZE
        }

        if extra_data:
            data.update(extra_data)

        self.test_cache = Cache.objects.create(**data)

    def _create_test_cache_partition(self):
        self.test_cache_partition = self.test_cache.partitions.create(
            name=TEST_CACHE_PARTITION_NAME
        )

    def _create_test_cache_partition_file(self, filename=None, file_size=None):
        cache_partition_file_total = len(self.test_cache_partition_files)

        file_size = file_size or TEST_CACHE_PARTITION_FILE_SIZE
        filename = filename or '{}_{}'.format(
            TEST_CACHE_PARTITION_FILE_FILENAME, cache_partition_file_total
        )

        with self.test_cache_partition.create_file(filename=filename) as file_object:
            file_object.write(
                force_bytes(s=' ' * file_size)
            )

        self.test_cache_partition_file = self.test_cache_partition.files.get(
            filename=filename
        )

        self.test_cache_partition_files.append(self.test_cache_partition_file)


class CacheViewTestMixin:
    def _request_test_cache_detail_view(self):
        return self.get(
            viewname='file_caching:cache_detail', kwargs={
                'cache_id': self.test_cache.pk
            }
        )

    def _request_test_cache_list_view(self):
        return self.get(viewname='file_caching:cache_list')

    def _request_test_cache_purge_view(self):
        return self.post(
            viewname='file_caching:cache_purge', kwargs={
                'cache_id': self.test_cache.pk
            }
        )

    def _request_test_cache_multiple_purge_view(self):
        return self.post(
            viewname='file_caching:cache_multiple_purge', data={
                'id_list': self.test_cache.pk
            }
        )


class FileCachingTaskTestMixin:
    def _execute_task_cache_partition_purge(self):
        task_cache_partition_purge.apply_async(
            kwargs={
                'cache_partition_id': self.test_cache_partition.pk
            }
        ).get()

    def _execute_task_cache_purge(self):
        task_cache_purge.apply_async(
            kwargs={
                'cache_id': self.test_cache.pk
            }
        ).get()



import errno
import os

import yaml

from mayan.apps.common.serialization import yaml_load

from .literals import (
    CONFIGURATION_FILENAME, CONFIGURATION_LAST_GOOD_FILENAME
)


class SettingNamespaceSingleton:
    """
    Self hosting bootstrap setting class.
    Allow managing setting in a compatible way before Mayan EDMS starts.
    """
    _setting_kwargs = {}
    _settings = {}

    class SettingNotFound(Exception):
        """Mostly a stand-in or typecast for KeyError for readability."""

    @classmethod
    def load_config_file(cls, filepath):
        try:
            with open(file=filepath) as file_object:
                file_object.seek(0, os.SEEK_END)
                if file_object.tell():
                    file_object.seek(0)
                    try:
                        return yaml_load(stream=file_object)
                    except yaml.YAMLError as exception:
                        exit(
                            'Error loading configuration file: {}; {}'.format(
                                filepath, exception
                            )
                        )
        except IOError as exception:
            if exception.errno == errno.ENOENT:
                # No config file, return empty dictionary
                return {}
            else:
                raise

    @classmethod
    def register_setting(cls, name, klass, kwargs=None):
        cls._settings[name] = klass
        cls._setting_kwargs[name] = kwargs or {}

    def __init__(self, global_symbol_table):
        self.global_symbol_table = global_symbol_table
        self.settings = {}
        for name, klass in self.__class__._settings.items():
            kwargs = self.__class__._setting_kwargs[name].copy()
            kwargs['name'] = name
            setting = klass(**kwargs)
            setting.namespace = self
            self.settings[name] = setting

    def get_config_file_setting(self, name):
        """
        Wrapper for load_config_file to cache the result.
        """
        if hasattr(self, '_cache_file_data'):
            file_data = self._cache_file_data
        else:
            filepath = self.get_setting_value(
                name='CONFIGURATION_FILEPATH'
            )

            file_data = self.load_config_file(filepath=filepath) or {}
            self._cache_file_data = file_data

        try:
            return file_data[name]
        except KeyError:
            raise SettingNamespaceSingleton.SettingNotFound

    def get_setting_value(self, name):
        """
        Wrapper that calls the individual setting .get_value method.
        Convenience method to allow returning setting values from the
        namespace.
        """
        try:
            return self.settings[name].get_value()
        except KeyError:
            raise SettingNamespaceSingleton.SettingNotFound

    def get_values(self, only_critical=False):
        """
        Return a dictionary will all the settings and their respective
        resolved values.
        """
        result = {}
        for name, setting in self.settings.items():
            # If only_critical is set to True load only the settings with
            # the critical flag. Otherwise load all.
            if only_critical and setting.critical or not only_critical:
                try:
                    result[name] = setting.get_value()
                except SettingNamespaceSingleton.SettingNotFound:
                    """
                    Not critical, we just avoid adding it to the result
                    dictionary.
                    """

        return result

    def update_globals(self, only_critical=False):
        """
        Insert all resolved values into the symbol table of the caller.
        """
        result = self.get_values(only_critical=only_critical)
        self.global_symbol_table.update(result)


class BaseSetting:
    def __init__(
        self, name, critical=False, has_default=False, default_value=None
    ):
        self.critical = critical
        self.name = name
        self.has_default = has_default
        self.default_value = default_value

    def _get_environment_value(self):
        return os.environ.get(self.get_environment_name())

    def get_default_value(self):
        return self.default_value

    def get_environment_name(self):
        return 'MAYAN_{}'.format(self.name)

    def get_value(self):
        """
        By default will try to get the value from the namespace symbol table,
        then the configuration file, and finally from the environment.
        """
        # Resolution order
        # 1 - Environment
        # 2 - Config
        # 3 - Global
        # 4 - Default
        try:
            return self.load_environment_value()
        except SettingNamespaceSingleton.SettingNotFound:
            try:
                return self.namespace.get_config_file_setting(name=self.name)
            except SettingNamespaceSingleton.SettingNotFound:
                try:
                    return self.namespace.global_symbol_table[self.name]
                except KeyError:
                    if self.has_default:
                        return self.get_default_value()
                    else:
                        raise SettingNamespaceSingleton.SettingNotFound

    def load_environment_value(self):
        value = self._get_environment_value()

        if value:
            try:
                return yaml_load(stream=value)
            except yaml.YAMLError as exception:
                exit(
                    'Error loading setting environment value: {}; {}'.format(
                        self.name, exception
                    )
                )
        else:
            raise SettingNamespaceSingleton.SettingNotFound


class FilesystemBootstrapSetting(BaseSetting):
    def __init__(self, name, critical=False, path_parts=None):
        self.critical = critical
        self.has_default = True
        self.name = name
        self.path_parts = path_parts

    def get_default_value(self):
        """
        The default value of this setting class is not static but calculated.
        """
        return os.path.join(
            # Can't use BASE_DIR from django.conf.settings
            # Use it from the global_symbol_table which should be the same
            self.namespace.global_symbol_table.get('BASE_DIR'), *self.path_parts
        )

    def get_value(self):
        """
        It is not possible to look for this setting in the config file
        because not even the config file setup has completed.
        This setting only supports being set from the environment.
        """
        try:
            return self.load_environment_value()
        except SettingNamespaceSingleton.SettingNotFound:
            if self.has_default:
                return self.get_default_value()
            else:
                raise SettingNamespaceSingleton.SettingNotFound


class MediaBootstrapSetting(FilesystemBootstrapSetting):
    def get_default_value(self):
        """
        The default value of this setting class is not static but calculated.
        """
        return os.path.join(
            self.namespace.get_setting_value(name='MEDIA_ROOT'),
            *self.path_parts
        )


def smart_yaml_load(value):
    if isinstance(value, dict):
        return value
    else:
        return yaml_load(
            stream=value or '{}',
        )


# FilesystemBootstrapSetting settings

SettingNamespaceSingleton.register_setting(
    name='CONFIGURATION_FILEPATH', klass=MediaBootstrapSetting,
    kwargs={
        'critical': True, 'path_parts': (CONFIGURATION_FILENAME,)
    }
)

# MediaBootstrapSetting settings

SettingNamespaceSingleton.register_setting(
    name='CONFIGURATION_LAST_GOOD_FILEPATH', klass=MediaBootstrapSetting,
    kwargs={
        'critical': True, 'path_parts': (CONFIGURATION_LAST_GOOD_FILENAME,)
    }
)
SettingNamespaceSingleton.register_setting(
    name='MEDIA_ROOT', klass=FilesystemBootstrapSetting, kwargs={
        'critical': True, 'path_parts': ('media',)
    }
)

# Normal settings

SettingNamespaceSingleton.register_setting(
    name='ALLOWED_HOSTS', klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': ['127.0.0.1', 'localhost', '[::1]']
    }
)
SettingNamespaceSingleton.register_setting(
    name='APPEND_SLASH', klass=BaseSetting,
)
SettingNamespaceSingleton.register_setting(
    name='AUTH_PASSWORD_VALIDATORS', klass=BaseSetting,
)
SettingNamespaceSingleton.register_setting(
    name='AUTHENTICATION_BACKENDS', klass=BaseSetting,
)
SettingNamespaceSingleton.register_setting(
    name='DATA_UPLOAD_MAX_MEMORY_SIZE', klass=BaseSetting,
)
SettingNamespaceSingleton.register_setting(
    name='DATABASES', klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': None
    }
)
SettingNamespaceSingleton.register_setting(
    name='DEBUG', klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': False
    }
)
SettingNamespaceSingleton.register_setting(
    name='DEFAULT_FROM_EMAIL', klass=BaseSetting,
)
SettingNamespaceSingleton.register_setting(
    name='DISALLOWED_USER_AGENTS', klass=BaseSetting,
)
SettingNamespaceSingleton.register_setting(
    name='EMAIL_BACKEND', klass=BaseSetting,
)
SettingNamespaceSingleton.register_setting(
    name='EMAIL_HOST', klass=BaseSetting,
)
SettingNamespaceSingleton.register_setting(
    name='EMAIL_HOST_PASSWORD', klass=BaseSetting,
)
SettingNamespaceSingleton.register_setting(
    name='EMAIL_HOST_USER', klass=BaseSetting,
)
SettingNamespaceSingleton.register_setting(
    name='EMAIL_PORT', klass=BaseSetting,
)
SettingNamespaceSingleton.register_setting(
    name='EMAIL_TIMEOUT', klass=BaseSetting,
)
SettingNamespaceSingleton.register_setting(
    name='EMAIL_USE_SSL', klass=BaseSetting,
)
SettingNamespaceSingleton.register_setting(
    name='EMAIL_USE_TLS', klass=BaseSetting,
)
SettingNamespaceSingleton.register_setting(
    name='FILE_UPLOAD_MAX_MEMORY_SIZE', klass=BaseSetting,
)
SettingNamespaceSingleton.register_setting(
    name='INTERNAL_IPS', klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': ['127.0.0.1']
    }
)
SettingNamespaceSingleton.register_setting(
    name='LOGIN_REDIRECT_URL', klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': 'common:home'
    }
)
SettingNamespaceSingleton.register_setting(
    name='LOGIN_URL', klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': 'authentication:login_view'
    }
)
SettingNamespaceSingleton.register_setting(
    name='LOGOUT_REDIRECT_URL', klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': 'authentication:login_view'
    }
)
SettingNamespaceSingleton.register_setting(
    name='LANGUAGES', klass=BaseSetting,
)
SettingNamespaceSingleton.register_setting(
    name='LANGUAGE_CODE', klass=BaseSetting,
)
SettingNamespaceSingleton.register_setting(
    name='SESSION_COOKIE_NAME', klass=BaseSetting,
)
SettingNamespaceSingleton.register_setting(
    name='SESSION_ENGINE', klass=BaseSetting,
)
SettingNamespaceSingleton.register_setting(
    name='STATIC_URL', klass=BaseSetting,
)
SettingNamespaceSingleton.register_setting(
    name='STATICFILES_STORAGE', klass=BaseSetting,
)
SettingNamespaceSingleton.register_setting(
    name='TIME_ZONE', klass=BaseSetting,
)
SettingNamespaceSingleton.register_setting(
    name='WSGI_APPLICATION', klass=BaseSetting,
)

# Celery

SettingNamespaceSingleton.register_setting(
    name='CELERY_TASK_ALWAYS_EAGER', klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': True
    }
)
SettingNamespaceSingleton.register_setting(
    name='CELERY_BROKER_LOGIN_METHOD', klass=BaseSetting
)
SettingNamespaceSingleton.register_setting(
    name='CELERY_BROKER_URL', klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': 'memory://'
    }
)
SettingNamespaceSingleton.register_setting(
    name='CELERY_BROKER_USE_SSL', klass=BaseSetting
)
SettingNamespaceSingleton.register_setting(
    name='CELERY_RESULT_BACKEND', klass=BaseSetting
)

# Mayan

SettingNamespaceSingleton.register_setting(
    name='COMMON_DISABLED_APPS', klass=BaseSetting, kwargs={
        'critical': True, 'has_default': True, 'default_value': ()
    }
)
SettingNamespaceSingleton.register_setting(
    name='COMMON_EXTRA_APPS', klass=BaseSetting, kwargs={
        'critical': True, 'has_default': True, 'default_value': ()
    }
)
SettingNamespaceSingleton.register_setting(
    name='DATABASE_ENGINE', klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': None
    }
)
SettingNamespaceSingleton.register_setting(
    name='DATABASE_NAME', klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': None
    }
)
SettingNamespaceSingleton.register_setting(
    name='DATABASE_USER', klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': None
    }
)
SettingNamespaceSingleton.register_setting(
    name='DATABASE_PASSWORD', klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': None
    }
)
SettingNamespaceSingleton.register_setting(
    name='DATABASE_HOST', klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': None
    }
)
SettingNamespaceSingleton.register_setting(
    name='DATABASE_PORT', klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': None
    }
)
SettingNamespaceSingleton.register_setting(
    name='DATABASE_CONN_MAX_AGE', klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': 0
    }
)
SettingNamespaceSingleton.register_setting(
    name='TESTING', klass=BaseSetting, kwargs={
        'has_default': True, 'default_value': False
    }
)



from django.utils.safestring import mark_safe

from mayan.apps.navigation.html_widgets import SourceColumnWidget

from .settings import (
    setting_preview_width, setting_preview_height, setting_thumbnail_width,
    setting_thumbnail_height
)


class ThumbnailWidget(SourceColumnWidget):
    template_name = 'documents/widgets/thumbnail.html'

    def get_extra_context(self):
        return {
            # Disable the clickable link if the document is in the trash
            'disable_title_link': self.value.is_in_trash,
            'gallery_name': 'document_list',
            'instance': self.value,
            'size_preview_width': setting_preview_width.value,
            'size_preview_height': setting_preview_height.value,
            'size_thumbnail_width': setting_thumbnail_width.value,
            'size_thumbnail_height': setting_thumbnail_height.value,
        }


def document_link(document):
    return mark_safe('<a href="%s">%s</a>' % (
        document.get_absolute_url(), document)
    )



from django.utils.translation import ugettext_lazy as _

from mayan.apps.navigation.classes import Link

from .icons import (
    icon_current_user_locale_profile_details,
    icon_current_user_locale_profile_edit,
)


link_current_user_locale_profile_details = Link(
    icon=icon_current_user_locale_profile_details,
    text=_('Locale profile'),
    view='locales:current_user_locale_profile_details'
)
link_current_user_locale_profile_edit = Link(
    icon=icon_current_user_locale_profile_edit,
    text=_('Edit locale profile'),
    view='locales:current_user_locale_profile_edit'
)



from django.contrib.contenttypes.models import ContentType

from rest_framework import serializers


class ContentTypeSerializer(serializers.ModelSerializer):
    class Meta:
        fields = ('app_label', 'id', 'model')
        model = ContentType



from ..classes import SettingNamespaceMigration


class TestNamespaceMigrationOne(SettingNamespaceMigration):
    def smart_settings_test_setting_0001(self, value):
        return '{}_0001'.format(value)


class TestNamespaceMigrationTwo(SettingNamespaceMigration):
    def smart_settings_test_setting_0001(self, value):
        return '{}_0001'.format(value)

    def smart_settings_test_setting_0002(self, value):
        return '{}_0002'.format(value)


class TestNamespaceMigrationInvalid(SettingNamespaceMigration):
    def smart_settings_test_setting(self, value):
        return 'invalid migration'


class TestNamespaceMigrationInvalidDual(SettingNamespaceMigration):
    def smart_settings_test_setting_with_longer_name(self, value):
        return 'invalid migration'



TEST_TAG_LABEL = 'test-tag'
TEST_TAG_LABEL_EDITED = 'test-tag-edited'
TEST_TAG_COLOR = '#001122'
TEST_TAG_COLOR_EDITED = '#221100'

TEST_TAG_INDEX_HAS_TAG = 'HAS_TAG'
TEST_TAG_INDEX_NO_TAG = 'NO_TAG'
TEST_TAG_INDEX_NODE_TEMPLATE = '''
{{% for tag in document.tags.all %}}
{{% if tag.label == "{tag_label}" %}}
{has_tag}
{{% else %}}
{no_tag}
{{% endif %}}
{{% empty %}}
{no_tag}
{{% endfor %}}
'''.format(
    tag_label='{}_0'.format(TEST_TAG_LABEL), has_tag=TEST_TAG_INDEX_HAS_TAG,
    no_tag=TEST_TAG_INDEX_NO_TAG
).replace('\n', '')



from django.shortcuts import get_object_or_404

from mayan.apps.acls.models import AccessControlList
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.documents.permissions import (
    permission_document_type_view, permission_document_view
)
from mayan.apps.documents.serializers.document_type_serializers import DocumentTypeSerializer
from mayan.apps.rest_api import generics
from mayan.apps.rest_api.api_view_mixins import ExternalObjectAPIViewMixin

from .models import SmartLink
from .permissions import (
    permission_smart_link_create, permission_smart_link_delete,
    permission_smart_link_edit, permission_smart_link_view
)
from .serializers import (
    ResolvedSmartLinkDocumentSerializer, ResolvedSmartLinkSerializer,
    SmartLinkConditionSerializer, SmartLinkDocumentTypeAddSerializer,
    SmartLinkDocumentTypeRemoveSerializer, SmartLinkSerializer
)


class APIResolvedSmartLinkDocumentListView(generics.ListAPIView):
    """
    get: Returns a list of the smart link documents that apply to the document.
    """
    mayan_object_permissions = {'GET': (permission_document_view,)}
    serializer_class = ResolvedSmartLinkDocumentSerializer

    def get_document(self):
        document = get_object_or_404(
            klass=Document, pk=self.kwargs['document_id']
        )

        AccessControlList.objects.check_access(
            obj=document, permissions=(permission_document_view,),
            user=self.request.user
        )

        return document

    def get_queryset(self):
        return self.get_smart_link().get_linked_document_for(
            document=self.get_document()
        )

    def get_serializer_context(self):
        """
        Extra context provided to the serializer class.
        """
        context = super().get_serializer_context()
        if self.kwargs:
            context.update(
                {
                    'document': self.get_document(),
                    'smart_link': self.get_smart_link()
                }
            )

        return context

    def get_smart_link(self):
        smart_link = get_object_or_404(
            klass=SmartLink.objects.get_for(document=self.get_document()),
            pk=self.kwargs['smart_link_id']
        )

        AccessControlList.objects.check_access(
            obj=smart_link, permissions=(permission_smart_link_view,),
            user=self.request.user
        )

        return smart_link


class APIResolvedSmartLinkView(generics.RetrieveAPIView):
    """
    get: Return the details of the selected resolved smart link.
    """
    lookup_url_kwarg = 'smart_link_id'
    mayan_object_permissions = {'GET': (permission_smart_link_view,)}
    serializer_class = ResolvedSmartLinkSerializer

    def get_document(self):
        document = get_object_or_404(
            klass=Document, pk=self.kwargs['document_id']
        )

        AccessControlList.objects.check_access(
            obj=document, permissions=(permission_document_view,),
            user=self.request.user
        )

        return document

    def get_serializer_context(self):
        """
        Extra context provided to the serializer class.
        """
        context = super().get_serializer_context()
        if self.kwargs:
            context.update(
                {
                    'document': self.get_document()
                }
            )

        return context

    def get_queryset(self):
        return SmartLink.objects.get_for(document=self.get_document())


class APIResolvedSmartLinkListView(generics.ListAPIView):
    """
    get: Returns a list of the smart links that apply to the document.
    """
    mayan_object_permissions = {'GET': (permission_smart_link_view,)}
    serializer_class = ResolvedSmartLinkSerializer

    def get_document(self):
        document = get_object_or_404(
            klass=Document, pk=self.kwargs['document_id']
        )

        AccessControlList.objects.check_access(
            obj=document, permissions=(permission_document_view,),
            user=self.request.user
        )

        return document

    def get_serializer_context(self):
        """
        Extra context provided to the serializer class.
        """
        context = super().get_serializer_context()
        if self.kwargs:
            context.update(
                {
                    'document': self.get_document()
                }
            )

        return context

    def get_queryset(self):
        return SmartLink.objects.filter(
            document_types=self.get_document().document_type
        )


class APISmartLinkConditionListView(
    ExternalObjectAPIViewMixin, generics.ListCreateAPIView
):
    """
    get: Returns a list of all the smart link conditions.
    post: Create a new smart link condition.
    """
    external_object_class = SmartLink
    external_object_pk_url_kwarg = 'smart_link_id'
    mayan_external_object_permissions = {
        'GET': (permission_smart_link_view,),
        'POST': (permission_smart_link_edit,)
    }
    ordering_fields = ('enabled',)
    serializer_class = SmartLinkConditionSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }

    def get_queryset(self):
        return self.external_object.conditions.all()

    def get_serializer_context(self):
        """
        Extra context provided to the serializer class.
        """
        context = super().get_serializer_context()
        if self.kwargs:
            context.update(
                {
                    'smart_link': self.external_object
                }
            )

        return context


class APISmartLinkConditionView(
    ExternalObjectAPIViewMixin, generics.RetrieveUpdateDestroyAPIView
):
    """
    delete: Delete the selected smart link condition.
    get: Return the details of the selected smart link condition.
    patch: Edit the selected smart link condition.
    put: Edit the selected smart link condition.
    """
    external_object_class = SmartLink
    external_object_pk_url_kwarg = 'smart_link_id'
    lookup_url_kwarg = 'smart_link_condition_id'
    mayan_external_object_permissions = {
        'DELETE': (permission_smart_link_edit,),
        'GET': (permission_smart_link_view,),
        'PATCH': (permission_smart_link_edit,),
        'PUT': (permission_smart_link_edit,)
    }
    serializer_class = SmartLinkConditionSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }

    def get_queryset(self):
        return self.external_object.conditions.all()

    def get_serializer_context(self):
        """
        Extra context provided to the serializer class.
        """
        context = super().get_serializer_context()
        if self.kwargs:
            context.update(
                {
                    'smart_link': self.external_object
                }
            )

        return context


class APISmartLinkListView(generics.ListCreateAPIView):
    """
    get: Returns a list of all the smart links.
    post: Create a new smart link.
    """
    mayan_object_permissions = {'GET': (permission_smart_link_view,)}
    mayan_view_permissions = {'POST': (permission_smart_link_create,)}
    queryset = SmartLink.objects.all()
    serializer_class = SmartLinkSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class APISmartLinkDetailView(generics.RetrieveUpdateDestroyAPIView):
    """
    delete: Delete the selected smart link.
    get: Return the details of the selected smart link.
    patch: Edit the selected smart link.
    put: Edit the selected smart link.
    """
    lookup_url_kwarg = 'smart_link_id'
    mayan_object_permissions = {
        'DELETE': (permission_smart_link_delete,),
        'GET': (permission_smart_link_view,),
        'PATCH': (permission_smart_link_edit,),
        'PUT': (permission_smart_link_edit,)
    }
    ordering_fields = ('dynamic_label', 'enabled', 'label')
    queryset = SmartLink.objects.all()
    serializer_class = SmartLinkSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class APISmartLinkDocumentTypeAddView(generics.ObjectActionAPIView):
    """
    post: Add a document type to a smart link.
    """
    lookup_url_kwarg = 'smart_link_id'
    mayan_object_permissions = {
        'POST': (permission_smart_link_edit,)
    }
    serializer_class = SmartLinkDocumentTypeAddSerializer
    queryset = SmartLink.objects.all()

    def object_action(self, request, serializer):
        document_type = serializer.validated_data['document_type']
        self.object._event_actor = self.request.user
        self.object.document_types_add(
            queryset=DocumentType.objects.filter(pk=document_type.pk)
        )


class APISmartLinkDocumentTypeListView(
    ExternalObjectAPIViewMixin, generics.ListAPIView
):
    """
    get: Return a list of the selected smart link document types.
    """
    external_object_class = SmartLink
    external_object_pk_url_kwarg = 'smart_link_id'
    mayan_external_object_permissions = {'GET': (permission_smart_link_view,)}
    mayan_object_permissions = {'GET': (permission_document_type_view,)}
    serializer_class = DocumentTypeSerializer

    def get_queryset(self):
        return self.external_object.document_types.all()


class APISmartLinkDocumentTypeRemoveView(generics.ObjectActionAPIView):
    """
    post: Remove a document type from a smart link.
    """
    lookup_url_kwarg = 'smart_link_id'
    mayan_object_permissions = {
        'POST': (permission_smart_link_edit,)
    }
    serializer_class = SmartLinkDocumentTypeRemoveSerializer
    queryset = SmartLink.objects.all()

    def object_action(self, request, serializer):
        document_type = serializer.validated_data['document_type']
        self.object._event_actor = self.request.user
        self.object.document_types_remove(
            queryset=DocumentType.objects.filter(pk=document_type.pk)
        )



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('document_states', '0016_auto_20191108_2332'),
    ]

    operations = [
        migrations.AlterField(
            model_name='workflow', name='label', field=models.CharField(
                help_text='A short text to describe the workflow.',
                max_length=255, unique=True, verbose_name='Label'
            ),
        ),
        migrations.AlterField(
            model_name='workflowstate', name='label', field=models.CharField(
                help_text='A short text to describe the workflow state.',
                max_length=255, verbose_name='Label'
            ),
        ),
        migrations.AlterField(
            model_name='workflowstateaction', name='label',
            field=models.CharField(
                help_text='A short text describing the action.',
                max_length=255, verbose_name='Label'
            ),
        ),
        migrations.AlterField(
            model_name='workflowtransition', name='label',
            field=models.CharField(
                help_text='A short text to describe the transition.',
                max_length=255, verbose_name='Label'
            ),
        ),
    ]



TEST_QUOTA_DOTTED_PATH = 'mayan.apps.quotas.tests.quota_backends.TestQuota'
TEST_QUOTA_WITH_MIXINS_DOTTED_PATH = 'mayan.apps.quotas.tests.quota_backends.TestQuotaWithMixins'
TEST_QUOTA_TEST_LIMIT = 1
TEST_QUOTA_TEST_LIMIT_EDITED = 2
TEST_QUOTA_DATA = {'test_limit': TEST_QUOTA_TEST_LIMIT}



import logging

from django.contrib import messages
from django.http import HttpResponseRedirect, JsonResponse
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.documents.literals import DOCUMENT_FILE_ACTION_PAGES_NEW
from mayan.apps.documents.models import Document, DocumentFile
from mayan.apps.documents.permissions import permission_document_file_new
from mayan.apps.documents.tasks import task_document_file_upload
from mayan.apps.storage.models import SharedUploadedFile

from ..exceptions import SourceException
from ..forms import NewFileForm, WebFormUploadForm, WebFormUploadFormHTML5
from ..models import SaneScanner, StagingFolderSource
from ..utils import get_upload_form_class

from .document_views import UploadBaseView

__all__ = ('DocumentFileUploadInteractiveView',)
logger = logging.getLogger(name=__name__)


class DocumentFileUploadInteractiveView(UploadBaseView):
    def create_source_form_form(self, **kwargs):
        return self.get_form_classes()['source_form'](
            prefix=kwargs['prefix'],
            source=self.source,
            show_expand=False,
            data=kwargs.get('data', None),
            files=kwargs.get('files', None),
        )

    def dispatch(self, request, *args, **kwargs):
        self.subtemplates_list = []

        self.document = get_object_or_404(
            klass=Document.valid, pk=kwargs['document_id']
        )

        AccessControlList.objects.check_access(
            obj=self.document, permissions=(permission_document_file_new,),
            user=self.request.user
        )

        try:
            DocumentFile.execute_pre_create_hooks(
                kwargs={
                    'document': self.document,
                    'shared_uploaded_file': None,
                    'user': self.request.user
                }
            )
        except Exception as exception:
            messages.error(
                message=_(
                    'Unable to upload new files for document '
                    '"%(document)s". %(exception)s'
                ) % {'document': self.document, 'exception': exception},
                request=self.request
            )
            return HttpResponseRedirect(
                redirect_to=reverse(
                    viewname='documents:document_file_list',
                    kwargs={'document_id': self.document.pk}
                )
            )

        self.tab_links = UploadBaseView.get_active_tab_links(
            document=self.document
        )

        return super().dispatch(request, *args, **kwargs)

    def forms_valid(self, forms):
        try:
            uploaded_file = self.source.get_upload_file_object(
                forms['source_form'].cleaned_data
            )
        except SourceException as exception:
            messages.error(message=exception, request=self.request)
        else:
            shared_uploaded_file = SharedUploadedFile.objects.create(
                file=uploaded_file.file
            )

            try:
                self.source.clean_up_upload_file(uploaded_file)
            except Exception as exception:
                messages.error(message=exception, request=self.request)

            if not self.request.user.is_anonymous:
                user = self.request.user
                user_id = self.request.user.pk
            else:
                user = None
                user_id = None

            try:
                DocumentFile.execute_pre_create_hooks(
                    kwargs={
                        'document': self.document,
                        'shared_uploaded_file': shared_uploaded_file,
                        'user': user
                    }
                )

                task_document_file_upload.apply_async(
                    kwargs={
                        'action': int(
                            forms['document_form'].cleaned_data.get('action')
                        ),
                        'comment': forms['document_form'].cleaned_data.get('comment'),
                        'document_id': self.document.pk,
                        'shared_uploaded_file_id': shared_uploaded_file.pk,
                        'user_id': user_id
                    }
                )
            except Exception as exception:
                message = _(
                    'Error executing document file upload task; '
                    '%(exception)s'
                ) % {
                    'exception': exception,
                }
                logger.critical(msg=message, exc_info=True)
                if self.request.is_ajax():
                    return JsonResponse(
                        data={'error': force_text(s=message)}, status=500
                    )
                else:
                    raise type(exception)(message)
            else:
                messages.success(
                    message=_(
                        'New document file queued for upload and will be '
                        'available shortly.'
                    ), request=self.request
                )

        return HttpResponseRedirect(
            redirect_to=reverse(
                viewname='documents:document_file_list', kwargs={
                    'document_id': self.document.pk
                }
            )
        )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(
            {
                'form_action': '{}?{}'.format(
                    reverse(
                        viewname=self.request.resolver_match.view_name,
                        kwargs=self.request.resolver_match.kwargs
                    ), self.request.META['QUERY_STRING']
                ),
                'object': self.document,
                'title': _(
                    'Upload a new file for document "%(document)s" '
                    'from source: %(source)s'
                ) % {'document': self.document, 'source': self.source.label},
                'submit_label': _('Submit')
            }
        )

        if not isinstance(self.source, StagingFolderSource) and not isinstance(self.source, SaneScanner):
            context.update(
                {
                    'form_css_classes': 'dropzone',
                    'form_disable_submit': True
                }
            )

        return context

    def get_document_form_initial(self):
        return {'action': DOCUMENT_FILE_ACTION_PAGES_NEW}

    def get_form_classes(self):
        source_form_class = get_upload_form_class(
            source_type_name=self.source.source_type
        )

        # Override source form class to enable the HTML5 file uploader
        if source_form_class == WebFormUploadForm:
            source_form_class = WebFormUploadFormHTML5

        return {
            'document_form': NewFileForm,
            'source_form': source_form_class
        }
